// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description: This file defines the classes that provide a database interface
//		 to the PBS server's accounting records.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------
#ifndef _PBS_ACCOUNTING_H_
#define _PBS_ACCOUNTING_H_ 1

#include <stdio.h>
#include <sys/param.h>
#include <sys/time.h>
#include <PBS_Hash.h>

// *****************************************************************************
// * class PBS_AccountingRecord:
// *	Each of these record represents a single line item in the PBS Server's
// *	accounting database.  Traditionally, these entries are located in files
// *	in the /usr/spool/PBS/server_priv/accounting directory.
// *****************************************************************************
class PBS_AccountingRecord
{
friend class PBS_Accounting;
friend class PBS_AccountingDatabase;

private:
	static time_t now;
	
	class PBS_NodeAccounting  ** nodeAcct;
	class PBS_UserAccounting   * userAcct;
	class PBS_QueueAccounting  * queueAcct;
	
	char     * job;	
	int        jobID;	// integer version of job
	char     * jobname;	// jobname=...
	char     * user;	// user=...
	char     * group;	// group=...
	char     * queue;	// queue=...
	time_t     ctime;	// ctime=#####
	time_t     qtime;	// qtime=#####
	time_t     etime;	// etime=#####
	time_t     start;	// start=#####
	time_t     end;		// end=#####
	int        exit_status;	// Exit_status=###
	time_t     cputime;	// resources_used.cput=##:##:##
	time_t     walltime;	// resources_used.walltime=##:##:##
	long       mem;		// resources_used.mem=#####kb
	long       vmem;	// resources_used.vmem=#####kb
	char     * exec_host;	// exec_host=hpc1b/1+hpc1b/0
	size_t     procCnt;	// computed from exec_host string
	char    ** nodeList;	// list of the nodes used by this job
	size_t     nodeCnt;	// number of nodes used by this job
	
	void clear            ( void );
	void nullify          ( char * &ptr );
	int  scanRecord       ( PBS_AccountingDatabase & db, const char * buf );
	void updateRecord     ( time_t endTime, time_t cpuTime, time_t wallTime );
	
	PBS_AccountingRecord  ( void );
	~PBS_AccountingRecord ( void );

public:
	static void startCycle ( void );
	static void endCycle   ( void );
	static PBS_AccountingRecord * create ( PBS_AccountingDatabase & db, const char * buf );
		
	inline const char          * getJob          ( void ) { return job;  }
	inline int                   getJobID        ( void ) { return jobID; }
	inline PBS_UserAccounting  * getUser         ( void ) { return userAcct; }
	inline PBS_QueueAccounting * getQueue        ( void ) { return queueAcct; }
	inline PBS_NodeAccounting ** getNodes        ( void ) { return nodeAcct; }
	inline const char          * getUserName     ( void ) { return user; }
	inline const char          * getQueueName    ( void ) { return queue; }
	inline const char          * getGroup        ( void ) { return group; }
	inline const char          * getJobName      ( void ) { return jobname; }
	inline time_t                getCreateTime   ( void ) { return ctime; }
	inline time_t                getQueueTime    ( void ) { return qtime; }
	inline time_t                getEligibleTime ( void ) { return etime; }
	inline time_t                getStartTime    ( void ) { return start;	}
	inline time_t                getEndTime      ( void ) { return end; }
	inline int                   getExitStatus   ( void ) { return exit_status; }
	inline time_t                getCpuTime      ( void ) { return cputime; }
	inline time_t                getWallTime     ( void ) { return walltime; }
	inline long                  getPhysMemory   ( void ) { return mem; }
	inline long                  getVirtMemory   ( void ) { return vmem; }
	inline const char          * getProcs        ( void ) { return exec_host; }
	inline size_t                getProcCnt      ( void ) { return procCnt; }
	inline const char         ** getNodeList     ( void ) { return (const char **)nodeList; }
	inline size_t                getNodeCnt      ( void ) { return nodeCnt; }
	
	size_t getNodeProcCnt ( const char * node );
	void   asciiDump      ( FILE * fp = stdout, char * prefix = NULL );

};


// *****************************************************************************
// * class PBS_Accounting:
// *	This object is used to store a collection of PBS_AccountingRecords
// *	that are associated with a specific PBS entity (such as a user or a 
// *	queue).
// *****************************************************************************
class PBS_Accounting
{
protected:
	PBS_AccountingRecord ** recs;
	size_t maxRecs;
	size_t numRecs;
	char  * name;
	
	ssize_t getIndex       ( const char * job );
		
public:
	PBS_Accounting  ( const char * Name );
	virtual ~PBS_Accounting ( void );
	
	PBS_AccountingRecord  * operator []  ( size_t idx );
	const char *            getName      ( void ) { return name; }
	size_t                  getCount     ( void ) { return numRecs; }
	PBS_AccountingRecord ** getRecords   ( void ) { return recs; }
	PBS_AccountingRecord  * find         ( const char * job );
	PBS_AccountingRecord  * attachRecord ( PBS_AccountingRecord * );
	PBS_AccountingRecord  * detachRecord ( PBS_AccountingRecord * );
	virtual void            asciiDump    ( FILE *fp=stdout, char * prefix=NULL);
	virtual void            summaryDump  ( const char * recType, FILE *fp=stdout, char*prefix=NULL);
};


// *****************************************************************************
// * class PBS_UserAccounting :
// *	This object is a subclass of the PBS_Accounting object and provides
// *	asciiDump and summaryDump methods consistent with User Accounting 
// *	records.
// *****************************************************************************
class PBS_UserAccounting : public PBS_Accounting
{
public:
	PBS_UserAccounting ( const char * Name ) : PBS_Accounting(Name) {}
	void asciiDump     ( FILE *fp=stdout, char * prefix=NULL);
};

// *****************************************************************************
// * class PBS_QueueAccounting :
// *	This object is a subclass of the PBS_Accounting object and provides
// *	asciiDump and summaryDump methods consistent with Queue Accounting 
// *	records.
// *****************************************************************************
class PBS_QueueAccounting : public PBS_Accounting
{
public:
	PBS_QueueAccounting ( const char * Name ) : PBS_Accounting(Name) {}
	void asciiDump      ( FILE *fp=stdout, char * prefix=NULL);
};

// *****************************************************************************
// * class PBS_NodeAccounting :
// *	This object is a subclass of the PBS_Accounting object and provides
// *	asciiDump and summaryDump methods consistent with Node Accounting 
// *	records.
// *****************************************************************************
class PBS_NodeAccounting : public PBS_Accounting
{
public:
	PBS_NodeAccounting ( const char * Name ) : PBS_Accounting(Name) {}
	void asciiDump     ( FILE *fp=stdout, char * prefix=NULL);
};



// *****************************************************************************
// * class PBS_AccountingDatabase :
// *	This object is the users primary interface to the PBS accounting 
// *	information.  At construction it will load all of the pertinent data
// *	files from the PBS accounting data path and will generate new 
// *	PBS_AccountingRecord objects tht represent each of them.  These objects
// *	will also be placed on a list by associated user and queue.  The objects
// *	in each of these lists are ordered by their numeric job name.
// *****************************************************************************
class PBS_AccountingDatabase : public PBS_Accounting
{
protected:
	time_t        daysOld;
	time_t        maxWallTime;
	char *        dataPath;
	char          dataFile[MAXPATHLEN];
	off_t         dataFileOffset;
	PBSStringHash nodes;
	PBSStringHash users;
	PBSStringHash queues;
	
public:
	void          initialize   ( void );
	void          purge        ( void );
	void          clear        ( void );
	int           scanRecord   ( const char * buf );

public:
	PBS_AccountingDatabase     ( const char * Name, time_t DaysOld=14, time_t MaxWallHours=200, const char * DataPath = NULL );
	~PBS_AccountingDatabase    ( void );

	void                  setThreshold ( time_t days ) { daysOld = days; }
	void                  update       ( void );
	void                  update       ( class PBS_Server & server );
	PBS_NodeAccounting  * findNode     ( const char * node );
	PBS_UserAccounting  * findUser     ( const char * user );
	PBS_QueueAccounting * findQueue    ( const char * queue );
	void                  asciiDump    ( FILE *fp = stdout, char * prefix = NULL);
};


#endif /* _PBS_ACCOUNTING_H_ */
