// *****************************************************************************
// * Copyright (c) 1999      Southeastern Universities Research Association,
// *                         Thomas Jefferson National Accelerator Facility
// * 
// * This software was developed under a United States Government license
// * described in the NOTICE file included as part of this distribution.
// *****************************************************************************
// *
// * File:        PBS_Job.cc
// * Author:      Walt Akers
// * Description: This file contains the source code for the functions defined
// *              in the PBS_Job class.
// *****************************************************************************
#include <PBS_Job.h>
#include <PBS_Server.h>

		
// *****************************************************************************
// * getTimeString :
// *	This is a static convenience method for obtaining a formatted time
// *	string from a time_t object.
// *****************************************************************************
static char * getTimeString(time_t t)
	{
	static char tempBuf[32];

	*tempBuf = 0;
	strftime(tempBuf, 32, "%m/%d/%Y %H:%M", (struct tm *)localtime(&t));
	
	return tempBuf;
	}

// *****************************************************************************
// * getDurationString :
// *	This is a static convenience method that converts a time_t into a 
// *	formatted duration.
// *****************************************************************************
static char * getDurationString ( time_t t )
	{
	static char tempBuf[32];
	*tempBuf = 0;
	sprintf(tempBuf, "%i:%02i:%02i", t/3600, (t%3600)/60, (t%60));
	
	return tempBuf;
	}
	
// *****************************************************************************
// * This static array of strings represents the names associated
// * with each of the job states.
// *****************************************************************************
char * PBS_Job::JobStateList[7] =
	{
	"EXITING",
	"HELD",
	"QUEUED",
	"RUNNING",
	"SUSPENDED",
	"TRANSIT",
	"WAITING"
	};
	
// *****************************************************************************
// * PBS_Job::Dependency::Dependency :
// *	This is the constructor for the internal Dependency class.  It 
// *	initializes all array pointers to NULL.
// *****************************************************************************
PBS_Job::Dependency::Dependency (void)
	: dataBuf    (NULL),
	  dataArray  (NULL),
	  after      (NULL),
	  afterok    (NULL),
	  afternotok (NULL),
	  afterany   (NULL),
	  before     (NULL),
	  beforeok   (NULL),
	  beforenotok(NULL),
	  beforeany  (NULL)
	{
	}
	
// *****************************************************************************
// * PBS_Job::Dependency::~Dependency :
// *	This is the destructor for the internal Dependency class.  It calls
// *	clear to free all memory that was allocated to the class and 
// *	restore the pointers to their default settings (NULL).
// *****************************************************************************
PBS_Job::Dependency::~Dependency (void)
	{
	clear();
	}

// *****************************************************************************
// * PBS_Job::Dependency::loadData :
// *	This method parses the string that contains the dependency information,
// *	and distributes it to the appropriate pointers.
// *****************************************************************************
int PBS_Job::Dependency::loadData   (char * str)
	{
	int    idx;
	char * ptr;
	
	// *********************************************************************
	// * Call clear to free any data that may currently be allocated.
	// *********************************************************************
	clear();
	
	// *********************************************************************
	// * Make a copy of the string and store it in the dataBuf variable, 
	// * this will be the source of all of your pointers.
	// *********************************************************************
	dataBuf = strdup(str);
	
	// *********************************************************************
	// * Count the number of commas and the number of colons.
	// * The commas+1 represent the number of NULL strings that will be
	// * required, and the colons represent the number of strings that 
	// * will be required.
	// *********************************************************************
	idx=1;
	for(ptr=dataBuf; *ptr; ptr++) if(*ptr==':' || *ptr==',') idx++;
	dataArray = (char **)malloc(idx*sizeof(char *));
	
	// *********************************************************************
	// * Now we'll walk through the databuf string, and everytime we find
	// * one of the dependency types that we're looking for we'll do three
	// * things.
	// *
	// * 1) Set that dependency types pointer to point to the current index
	// *    in the dataArray.
	// * 2) Copy all of the subsequent entries into the dataArray starting
	// *    at the current index.
	// * 3) Set the next entry in the dataArray to NULL to indicate the
	// *    end of that list.
	// *********************************************************************
	ptr = dataBuf;
	idx = 0;
	while(*ptr)
		{
		int found = 0;
		
		while(*ptr && !found)
			{
			found = 1;
			switch(*ptr)
				{
				case 'a':
				if     (!strncmp(ptr, "after:", 6)) 
					{
					after = dataArray+idx;
					ptr+=6;
					}
				else if(!strncmp(ptr, "afterok:", 8))
					{
					afterok = dataArray+idx;
					ptr+=8;
					}
				else if(!strncmp(ptr, "afternotok:", 11))
					{
					afternotok = dataArray+idx;
					ptr+=11;
					}
				else if(!strncmp(ptr, "afterany:",  9))
					{
					afterany = dataArray+idx;
					ptr+=11;
					}
				else	{
					while(*ptr && *ptr!=',') ptr++;
					if(*ptr == ',') ptr++;
					found = 0;
					}
				break;
			
				case 'b':
				if     (!strncmp(ptr, "before:", 7))
					{
					before = dataArray+idx;
					ptr+=7;
					}
				else if(!strncmp(ptr, "beforeok:", 9))
					{
					beforeok = dataArray+idx;
					ptr+=9;
					}
				else if(!strncmp(ptr, "beforenotok:", 12))
					{
					beforenotok = dataArray+idx;
					ptr+=12;
					}
				else if(!strncmp(ptr, "beforeany:", 10))
					{
					beforeany = dataArray+idx;
					ptr+=10;
					}
				else	{
					while(*ptr && *ptr!=',') ptr++;
					if(*ptr == ',') ptr++;
					found = 0;
					}
				break;
				
				default:
				while(*ptr && *ptr!=',') ptr++;
				if(*ptr == ',') ptr++;
				found = 0;
				break;
				}
			}

		int done = 0;
		while(*ptr && !done)
			{
			char * ePtr = ptr;
			while(*ePtr && *ePtr!=':' && *ePtr!=',') ePtr++;
			if(*ePtr!=':') done = 1;
			if(*ePtr) *(ePtr++)=0;
			dataArray[idx++] = ptr;
			if(done) dataArray[idx++]=NULL;
			ptr = ePtr;
			}
		}
	return idx;
	}
	
	
// *****************************************************************************
// * PBS_Job::Dependency::clear :
// *	This method frees all memory that was previously allocated by the
// *	class and then sets the pointers to their default value (NULL).
// *****************************************************************************
void PBS_Job::Dependency::clear (void)
	{
	free (dataBuf);
	free (dataArray);

	dataBuf     = NULL;
	dataArray   = NULL;
	after       = NULL;
	afterok     = NULL;
	afternotok  = NULL;
	afterany    = NULL;
	before      = NULL;
	beforeok    = NULL;
	beforenotok = NULL;
	beforeany   = NULL;
	}


// *****************************************************************************
// * PBS_Job::readHoldTypeValue :
// *	This method will translate the Hold_Type character string attribute
// *	into an integer value.
// *****************************************************************************
void PBS_Job::readHoldTypeValue (int & val, struct attrl * attrib)
	{
	char* ptr = attrib->value;

	val = HOLD_NONE;
	
	while(ptr && *ptr)
		{
		switch(tolower(*ptr))
			{
			case 'u':
			val |= HOLD_USER;
			break;
			
			case 's':
			val |= HOLD_SYSTEM;
			break;
			
			case 'o':
			val |= HOLD_OTHER;
			break;
			}
		ptr++;
		}	
	}
	
// *****************************************************************************
// * PBS_Job::readMailPointValue :
// *	This method will translate the Mail_Points character string attribute
// *	into an integer value.
// *****************************************************************************
void PBS_Job::readMailPointValue (int & val, struct attrl * attrib)
	{
	char* ptr = attrib->value;

	val = MAIL_DEFAULT;
	
	while(ptr && *ptr)
		{
		switch(tolower(*ptr))
			{
			case 'a':
			val |= MAIL_ON_ABORT;
			break;
			
			case 'b':
			val |= MAIL_ON_BEGIN;
			break;
			
			case 'e':
			val |= MAIL_ON_EXIT;
			break;
			}
		ptr++;
		}	
	}
	
// *****************************************************************************
// * PBS_Job::readJobStateValue :
// *	This method will translate the Job_State character string attribute
// *	into an integer value.
// *****************************************************************************
void PBS_Job::readJobStateValue (JobState & val, struct attrl * attrib)
	{
	val = QUEUED;
	
	switch(tolower(*attrib->value))
		{
		case 'e':
		val = EXITING;
		break;
		
		case 'h':
		val = HELD;
		break;
		
		case 'q':
		val = QUEUED;
		break;
		
		case 'r':
		val = RUNNING;
		break;
		
		case 's':
		val = SUSPENDED;
		break;
		
		case 't':
		val = TRANSIT;
		break;
		
		case 'w':
		val = WAITING;
		break;
		}
	}

// *****************************************************************************
// * PBS_Job::loadData :
// *	This method can be called by the constructor or independently to cause
// *	the object to be populated with new job data.
// *
// *	The batch_status object is retrieved by calling the pbs_statjob function
// *	using the connection ID of the server.  Each attribute of that object
// *	is then examined and the value is loaded into this class.
// *****************************************************************************
int PBS_Job::loadData ( struct batch_status * job )
	{
	struct attrl        * attrib;

	// *********************************************************************
	// * Call clear to clear the contents of the object.
	// *********************************************************************
	clear();

	// *********************************************************************
	// * Copy the name of the job into the name variable.
	// *********************************************************************
	name = strdup(job->name);

	// *********************************************************************
	// * Process the contents of each of the attributes that was returned
	// * as part of the batch_status object.
	// *********************************************************************
	for(attrib=job->attribs; attrib!=NULL; attrib = attrib->next)
		{
		switch(tolower(*attrib->name))
			{
			case 'a':
			if(!strcmp(attrib->name, "Account_Name")) readStringValue(account_name, attrib);
			else LogError("PBS_Job::loadData", "Didn't process attribute %s", attrib->name);
			break;
			
			case 'c':
			if     (!strcmp(attrib->name, "Checkpoint")) readStringValue(checkpoint, attrib);
			else if(!strcmp(attrib->name, "comment")) readStringValue(comment, attrib);
			else if(!strcmp(attrib->name, "ctime"))   readIntegerValue(ctime, attrib);
			else LogError("PBS_Job::loadData", "Didn't process attribute %s", attrib->name);
			break;
			
			case 'd':
			if(!strcmp(attrib->name, "depend")) depend.loadData(attrib->value);
			else LogError("PBS_Job::loadData", "Didn't process attribute %s", attrib->name);
			break;
			
			case 'e':
			if     (!strcmp(attrib->name, "Error_Path")) readStringValue(error_path, attrib);
			else if(!strcmp(attrib->name, "Execution_Time")) readIntegerValue(execution_time, attrib);
			else if(!strcmp(attrib->name, "etime")) readIntegerValue(etime, attrib);
			else if(!strcmp(attrib->name, "exec_host")) readListValue(exec_host, attrib, '+');
			else if(!strcmp(attrib->name, "egroup")) readStringValue(egroup, attrib);
			else if(!strcmp(attrib->name, "euser")) readStringValue(euser, attrib);
			else LogError("PBS_Job::loadData", "Didn't process attribute %s", attrib->name);
			break;
			
			case 'g': 
			if     (!strcmp(attrib->name, "group_list")) readListValue(group_list, attrib);
			else LogError("PBS_Job::loadData", "Didn't process attribute %s", attrib->name);
			break;
			
			case 'h':
			if     (!strcmp(attrib->name, "Hold_Types")) readHoldTypeValue(hold_type, attrib);
			else if(!strcmp(attrib->name, "hashname")) readStringValue(hashname, attrib);
			else LogError("PBS_Job::loadData", "Didn't process attribute %s", attrib->name);
			break;
			
			case 'i':
			if     (!strcmp(attrib->name, "interactive")) readBooleanValue(interactive, attrib);
			else LogError("PBS_Job::loadData", "Didn't process attribute %s", attrib->name);
			break;

			case 'j':
			if     (!strcmp(attrib->name, "Job_Name")) readStringValue(job_name, attrib);
			else if(!strcmp(attrib->name, "Join_Path")) readBooleanValue(join_path, attrib);
			else if(!strcmp(attrib->name, "Job_Owner"))
				{
				readStringValue(job_owner, attrib);
				job_owner_id = getUserIDFromName(job_owner);
				}
			else if(!strcmp(attrib->name, "job_state")) readJobStateValue(job_state, attrib);
			else LogError("PBS_Job::loadData", "Didn't process attribute %s", attrib->name);
			break;
			
			case 'k':
			if     (!strcmp(attrib->name, "Keep_Files")) readIntegerValue(keep_files, attrib);
			else LogError("PBS_Job::loadData", "Didn't process attribute %s", attrib->name);
			break;
			
			case 'm':
			if     (!strcmp(attrib->name, "Mail_Points")) readMailPointValue(mail_points, attrib);
			else if(!strcmp(attrib->name, "Mail_Users")) readListValue(mail_users, attrib);
			else if(!strcmp(attrib->name, "mtime")) readIntegerValue(mtime, attrib);
			else LogError("PBS_Job::loadData", "Didn't process attribute %s", attrib->name);
			break;
			
			case 'o':
			if     (!strcmp(attrib->name, "Output_Path")) readStringValue(output_path, attrib);
			else LogError("PBS_Job::loadData", "Didn't process attribute %s", attrib->name);
			break;
			
			case 'p':
			if     (!strcmp(attrib->name, "Priority")) readIntegerValue(priority, attrib);
			else LogError("PBS_Job::loadData", "Didn't process attribute %s", attrib->name);
			break;
			
			case 'q':
			if     (!strcmp(attrib->name, "qtime")) readIntegerValue(qtime, attrib);
			else if(!strcmp(attrib->name, "queue"))
				{
				readStringValue(queue_name, attrib);
				queue_id = getQueueIDFromName(queue_name);
				}
			else if(!strcmp(attrib->name, "queue_rank")) readIntegerValue(queue_rank, attrib);
			else if(!strcmp(attrib->name, "queue_type")) readQueueTypeValue(queue_type, attrib);
			else LogError("PBS_Job::loadData", "Didn't process attribute %s", attrib->name);
			break;
			
			case 'r':
			if     (!strcmp(attrib->name, "Rerunable")) readBooleanValue(rerunable, attrib);
			else if(!strcmp(attrib->name, "Resource_List")) readResourceValue(resource_list, attrib);
			else if(!strcmp(attrib->name, "resources_used")) readResourceValue(resources_used, attrib);
			else LogError("PBS_Job::loadData", "Didn't process attribute %s", attrib->name);
			break;
			
			case 's':
			if     (!strcmp(attrib->name, "Shell_Path_List")) readListValue(shell_path_list, attrib);
			else if(!strcmp(attrib->name, "stagein")) readListValue(stage_in, attrib);
			else if(!strcmp(attrib->name, "stageout")) readListValue(stage_out, attrib); 
			else if(!strcmp(attrib->name, "session_id")) readIntegerValue(session_id, attrib);
			else if(!strcmp(attrib->name, "server")) break;
			else if(!strcmp(attrib->name, "substate")) break;
			else LogError("PBS_Job::loadData", "Didn't process attribute %s", attrib->name);
			break;
			
			case 'u':
			if     (!strcmp(attrib->name, "User_List")) readListValue(user_list, attrib);
			else LogError("PBS_Job::loadData", "Didn't process attribute %s", attrib->name);
			break;
			
			case 'v':
			if     (!strcmp(attrib->name, "Variable_List")) readListValue(variable_list, attrib);
			else LogError("PBS_Job::loadData", "Didn't process attribute %s", attrib->name);
			break;
			
			default:
			LogError("PBS_Job::loadData", "Didn't process attribute %s", attrib->name);
			break;
			}
		}

	// *********************************************************************
	// Here we will compute the duration of a running job. The
	// duration of the job will be estimated based on the user's specified 
	// walltime and the historical cpu time of THIS job. The most 
	// restrictive of these terms will be used as the duration.
	// 	
	// In the event that a job has not started and there is no historical
	// cpu time versus walltime for this estimate, then the job will be 
	// considered 100% efficient.
	// *********************************************************************
	char * str     = NULL;
	double wallReq = 0.0;
	double cpuReq  = 0.0;
	duration       = 0;
	
	if((str = resource_list.find("walltime")) && *str)
		wallReq = stringToDuration(str, 0);
	if((str = resource_list.find("cput")) && *str)
		cpuReq  = stringToDuration(str, 0);
	
	if(wallReq || cpuReq)
		{
		double wallUse = 0.0;
		double wallDur = 0.0;
		double cpuUse  = 0.0;
		double cpuDur  = 0.0;
		duration       = 0;

		if((str = resources_used.find("walltime")) && *str)
			wallUse = stringToDuration(str, 0);
		if((str = resources_used.find("cput")) && *str)
			cpuUse = stringToDuration(str, 0);

		wallDur = wallReq;
		cpuDur  = cpuReq * ((wallUse>0)?wallUse:1)/((cpuUse>0)?cpuUse:1);		

		if(wallReq>0 && cpuReq>0) duration = (time_t)((wallDur<cpuDur)?wallDur:cpuDur);
		else if(wallReq>0)        duration = (time_t)wallDur;
		else                      duration = (time_t)cpuDur;
		}
	if(!duration) duration = DEFAULT_DURATION;
	
	// *********************************************************************
	// Automatically calculate the startTime variable using the
	// system provided 'walltime' used resource.  If the job has not been
	// started then the startTime should be zero.
	// *********************************************************************
	startTime = 0;
	if(job_state==RUNNING || job_state==EXITING)
		{
		str = resources_used.find("walltime");
		time_t wallTimeUsed = 0;
	
		if(str && *str)
			{
			int multx = 1;
			for(const char *ptr=str; multx<3600 && (ptr=strchr(ptr, ':'))!=NULL; ptr++)
				{
				multx*=60;
				}
	
			wallTimeUsed    = 0;
			while(*str && multx)
				{
				wallTimeUsed += atoi(str)*multx;
				if(*(str+=strcspn(str, ":"))) str++;
				multx/=60;
				}
			}
		
		if(str && wallTimeUsed < server.getLoadDataTime())
			{
			startTime = server.getLoadDataTime() - wallTimeUsed;
			}
		else if(mtime != 0) startTime = mtime;
		else startTime = server.getLoadDataTime();
		}

	// *********************************************************************
	// * Automatically calculate the number of processors requested in the
	// * resources. This is an advisory value that is required by other 
	// * system components.
	// *********************************************************************
	str = resource_list.find("nodes");
	if(str && *str)
		{
		char * buf     = strdup(str);
		char * ptr     = buf;
		char * nextPtr = buf;
	
		while(ptr && *ptr)
			{
			size_t nodeCnt = 1;
			size_t ppnCnt  = 1;

			if((nextPtr = strchr(ptr, '+'))!=NULL) *nextPtr = 0;
			if(isdigit(*ptr)) nodeCnt = atol(ptr);
			if((ptr = strstr(ptr, "ppn="))!=NULL) ppnCnt = atol(ptr+4);
			proc_cnt += nodeCnt*ppnCnt;
			if(nextPtr)
				{
				*nextPtr = '+';
				ptr = nextPtr+1;
				}
			else	ptr = NULL;
			}
		free(buf);
		}
	else if((str = resource_list.find("nodect"))!=NULL)
		{
		proc_cnt = atoi(str);
		}
	if(!proc_cnt) proc_cnt = 1;

	return 0;
	}

// *****************************************************************************
// * PBS_Job::clear :
// *	This method releases all memory that was allocated to the object and 
// *	then restores all internal variables to their default values.  This 
// *	method is generally called as part of object destruction, or to clear
// *	values so they can be repopulated from the server.
// *****************************************************************************
void PBS_Job::clear ( void )
	{
	nullify(name);
	nullify(account_name);
	nullify(checkpoint);
	nullify(error_path);
	nullify(group_list);
	nullify(job_name);
	nullify(mail_users);
	nullify(output_path);
	nullify(shell_path_list);
	nullify(stage_in);
	nullify(stage_out);
	nullify(user_list);
	nullify(variable_list);
	nullify(comment);
	nullify(exec_host);
	nullify(egroup);
	nullify(euser);
	nullify(hashname);
	nullify(job_owner);
	job_owner_id = -1;	
	nullify(queue_name);
	queue_id = -1;
					
	depend.clear();
	resource_list.clear();
	resources_used.clear();
	
	execution_time = 0;
	hold_type      = HOLD_NONE;
	join_path      = FALSE;
	keep_files     = KEEP_NOTHING;
	mail_points    = MAIL_DEFAULT;
	priority       = 0;
	rerunable      = TRUE;
	ctime          = 0;
	etime          = 0;
	duration       = DEFAULT_DURATION;
	mtime          = 0;
	qtime          = 0;
	interactive    = FALSE;
	job_state      = QUEUED;
	queue_rank     = 0; 
	session_id     = 0;
	startTime      = 0;
	weight         = 1.0;
	proc_cnt       = 0;
	}


// *****************************************************************************
// * PBS_Job::asciiDump :
// *	This method generates a plain text report of the contents of the
// *	object.  This method is generally used for debugging, however, the
// *	information is also useful for reporting and verifying changes made
// *	from other sources.  
// *****************************************************************************
void PBS_Job::asciiDump ( FILE * fp )
	{
	int  i;
	
	fprintf(fp, "-----------------------------------------------------\n");
	fprintf(fp, "JOB %s\n", name);
	fprintf(fp, "-----------------------------------------------------\n");
	fprintf(fp, "JOB DATA:\n");
	fprintf(fp, "     Server:        %s\n", server.getName());
	fprintf(fp, "     Name:          %s\n", job_name?job_name:name);
	fprintf(fp, "     Account:       %s\n", account_name?account_name:"\0");
	fprintf(fp, "     Session ID:    %i\n", session_id);
	fprintf(fp, "     State:         %s\n", JobStateList[job_state]);
	fprintf(fp, "     Priority:      %i\n", priority);
	fprintf(fp, "     Delayed Until: %s\n", execution_time?getTimeString(execution_time):"\0"); 
	fprintf(fp, "     Created:       %s\n", ctime?getTimeString(ctime):"\0");
	fprintf(fp, "     Enqueued:      %s\n", qtime?getTimeString(qtime):"\0");
	fprintf(fp, "     Last Modified: %s\n", mtime?getTimeString(mtime):"\0");
	fprintf(fp, "     Eligible Time: %s\n", etime?getTimeString(etime):"\0");
	fprintf(fp, "     Start Time:    %s\n", startTime?getTimeString(startTime):"\0");
	fprintf(fp, "     Duration:      %s\n", duration?getDurationString(duration):"\0");
	fprintf(fp, "     Rerunable:     %s\n", rerunable?"Yes":"No");
	fprintf(fp, "     Interactive:   %s\n", interactive?"Yes":"No");
	fprintf(fp, "     Check Point:   %s\n", checkpoint?checkpoint:"\0");
	fprintf(fp, "     Comment:       %s\n", comment?comment:"\0");
	fprintf(fp, "     Local Weight:  %g\n", weight);
	fprintf(fp, "     Proccesor Cnt: %i\n", proc_cnt);
			
	fprintf(fp, "\nQUEUE DATA:\n");
	fprintf(fp, "     Queue Name:    %s\n", queue_name);
	fprintf(fp, "     Queue Type:    %s\n", queue_type==EXECUTION?"Execution":"Routing");
	fprintf(fp, "     Local QueueID: %i\n", queue_id);
	fprintf(fp, "     Queue Rank:    %i\n", queue_rank);
	fprintf(fp, "     Hold Types:    %s%s%s\n", 
		(hold_type&HOLD_USER)?"USER ":"\0",
		(hold_type&HOLD_SYSTEM)?"SYSTEM ":"\0",
		(hold_type&HOLD_OTHER)?"OTHER":"\0");
	
	fprintf(fp, "\nUSER DATA:\n");
	fprintf(fp, "     Owner:         %s\n", job_owner?job_owner:"\0");
	fprintf(fp, "     Local OwnerID: %i\n", job_owner_id);
	fprintf(fp, "     Effective UID: %s\n", euser?euser:"\0"); 
	fprintf(fp, "     Effective GID: %s\n", egroup?egroup:"\0");
	fprintf(fp, "     User List:");
	for(i=0; user_list && user_list[i]!=NULL; i++)
		fprintf(fp, "%*c%s\n", (i?20:5), 0x20, user_list[i]);
	if(i==0) fprintf(fp, "\n");
	fprintf(fp, "     Group List:");
	for(i=0; group_list && group_list[i]!=NULL; i++)
		fprintf(fp, "%*c%s\n", (i?20:4), 0x20, group_list[i]);
	if(i==0) fprintf(fp, "\n");
	
	fprintf(fp, "\nNODE DATA:\n");
	fprintf(fp, "     Executing On:");
	for(i=0; exec_host && exec_host[i]!=NULL; i++)
		fprintf(fp, "%*c%s\n", (i?20:2), 0x20, exec_host[i]);
	if(!i) fprintf(fp, "\n");
	
	fprintf(fp, "\nINPUT/OUTPUT DATA:\n");
	fprintf(fp, "     Error Path:    %s\n", error_path?error_path:"\0");
	fprintf(fp, "     Output Path:   %s\n", output_path?output_path:"\0");
	fprintf(fp, "     Join I/O Path: %s\n", join_path?"Yes":"No");
	fprintf(fp, "     Keep Files:    %s%s\n", 
		(keep_files&KEEP_STDERR)?"STDERR ":"\0",
		(keep_files&KEEP_STDOUT)?"STDOUT":"\0");
	fprintf(fp, "     Hash Name:     %s\n", hashname?hashname:"\0");
	fprintf(fp, "     Copy In File:");
	for(i=0; stage_in && stage_in[i]!=NULL; i++)
		fprintf(fp, "%*c%s\n", (i?20:2), 0x20, stage_in[i]);
	if(!i) fprintf(fp, "\n");
	fprintf(fp, "     Copy Out File:");
	for(i=0; stage_out && stage_out[i]!=NULL; i++)
		fprintf(fp, "%*c%s\n", (i?20:1), 0x20, stage_out[i]);
	if(!i) fprintf(fp, "\n");
	fprintf(fp, "     Shell List:");
	for(i=0; shell_path_list && shell_path_list[i]!=NULL; i++)
		fprintf(fp, "%*c%s\n", (i?20:4), 0x20, shell_path_list[i]);
	if(!i) fprintf(fp, "\n");
	fprintf(fp, "     Variable List:");
	for(i=0; variable_list && variable_list[i]!=NULL; i++)
		fprintf(fp, "%*c%s\n", (i?20:1), 0x20, variable_list[i]);
	if(!i) fprintf(fp, "\n");
	
	fprintf(fp, "\nMAIL DATA:\n");
	fprintf(fp, "     Mail Points    %s%s%s\n",
		(mail_points&MAIL_ON_ABORT)?"ABORT ":"\0",
		(mail_points&MAIL_ON_BEGIN)?"START ":"\0",
		(mail_points&MAIL_ON_EXIT)?"EXIT":"\0");
	fprintf(fp, "     Mail Users:");
	for(i=0; mail_users && mail_users[i]!=NULL; i++)
		fprintf(fp, "%*c%s\n", (i?20:4), 0x20, mail_users[i]);
	if(!i) fprintf(fp, "\n");

	fprintf(fp, "\nDEPENDENCIES:\n");
	if(depend.after)
		{
		fprintf(fp, "     AFTER:");
		for(i=0; depend.after[i]!=NULL; i++) 
			fprintf(fp, "%*c%s\n", (i?20:9), 0x20, depend.after[i]);
		}
	if(depend.afterok)
		{
		fprintf(fp, "     AFTER OK:");
		for(i=0; depend.afterok[i]!=NULL; i++) 
			fprintf(fp, "%*c%s\n", (i?20:6), 0x20, depend.afterok[i]);
		}
	if(depend.afternotok)
		{
		fprintf(fp, "     AFTER NOT OK:");
		for(i=0; depend.afternotok[i]!=NULL; i++) 
			fprintf(fp, "%*c%s\n", (i?20:2), 0x20, depend.afternotok[i]);
		}
	if(depend.afterany)
		{
		fprintf(fp, "     AFTER ANY:");
		for(i=0; depend.afterany[i]!=NULL; i++) 
			fprintf(fp, "%*c%s\n", (i?20:5), 0x20, depend.afterany[i]);
		}
	if(depend.before)
		{
		fprintf(fp, "     BEFORE:");
		for(i=0; depend.before[i]!=NULL; i++) 
			fprintf(fp, "%*c%s\n", (i?20:8), 0x20, depend.before[i]);
		}
	if(depend.beforeok)
		{
		fprintf(fp, "     BEFORE OK:");
		for(i=0; depend.beforeok[i]!=NULL; i++) 
			fprintf(fp, "%*c%s\n", (i?20:5), 0x20, depend.beforeok[i]);
		}
	if(depend.beforenotok)
		{
		fprintf(fp, "     BEFORE NOT OK:");
		for(i=0; depend.beforenotok[i]!=NULL; i++) 
			fprintf(fp, "%*c%s\n", (i?20:1), 0x20, depend.beforenotok[i]);
		}
	if(depend.beforeany)
		{
		fprintf(fp, "     BEFORE ANY:");
		for(i=0; depend.beforeany[i]!=NULL; i++) 
			fprintf(fp, "%*c%s\n", (i?20:4), 0x20, depend.beforeany[i]);
		}
	
	fprintf(fp, "\nRESOURCES\n");
	fprintf(fp, "      Resources Requested:\n");
	resource_list.asciiDump(fp, "                    ");
	fprintf(fp, "\n      Resources Used:\n");
	resources_used.asciiDump(fp, "                    ");
	
	fprintf(fp, "\n");
	fflush (fp);
	}
	
// *****************************************************************************
// * PBS_JobList::PBS_JobList :
// *	Default constructor for the class.  It initializes all internal
// *	variables to their default values.
// *****************************************************************************
PBS_JobList::PBS_JobList ( void )
	: jobs(NULL),
	  jobCnt(0)
	{
	}

// *****************************************************************************
// * PBS_JobList::PBS_JobList :
// *	Parameterized constructor for the class.  It initializes the
// *	internal variables using an array of jobs that is terminated
// *	by a NULL pointer.
// *
// *	The array becomes the property of the class and should not be
// *	accessed directly by the caller.
// *****************************************************************************
PBS_JobList::PBS_JobList ( PBS_Job ** joblist )
	: jobs(NULL),
	  jobCnt(0)
	{
	addJobs ( joblist );
	}

// *****************************************************************************
// * PBS_JobList::~PBS_JobList :
// *	Destructor for the object.  It calls clear to free the array structure
// *	associated with the object. The PBS_Server is responsible for 
// *	deleteing the individual PBS_Jobs.
// *****************************************************************************
PBS_JobList::~PBS_JobList ( void )
	{
	if(jobs) delete[] jobs;
	jobs   = NULL;
	jobCnt = 0;
	}

// *****************************************************************************
// * PBS_JobList::addJobs :
// *	This method populates the object using an external array of 
// *	jobs that is termimated by a NULL pointer.  The array becomes 
// *	the property of the class and should not be accessed directly 
// *	by the caller.
// *****************************************************************************
void PBS_JobList::addJobs ( PBS_Job ** joblist )
	{
	if(jobs) delete[] jobs;
	jobs   = joblist;
	jobCnt = 0;
	
	if(jobs!=NULL) while(jobs[jobCnt]!=NULL) jobCnt++;
	}
		
// *****************************************************************************
// * PBS_JobList::getJob :
// *	This method returns a specific job by name.  If the job doesn't
// *	exist in the list, then a NULL will be returned.
// *****************************************************************************
PBS_Job * PBS_JobList::getJob ( char * name )
	{
	PBS_Job * result = NULL;
	int idx;
		
	if(name && *name)
		{
		for(idx=0; idx<jobCnt && strcmp(jobs[idx]->getName(), name); idx++);
		if(idx<jobCnt) result = jobs[idx];
		}
	
	return result;
	}
	
// *****************************************************************************
// * PBS_JobList::getJobCount :
// *	Returns a count of all jobs that are currently in the list.
// *****************************************************************************
size_t PBS_JobList::getJobCount ( void )
	{
	return jobCnt;
	}

// *****************************************************************************
// * PBS_JobList::getJobs :
// *	This method returns an array of all jobs that are available in 
// *	the list.  The user is responsible for deleteing the array 
// *	that is allocated by this call.
// *****************************************************************************
PBS_Job ** PBS_JobList::getJobs ( void )
	{
	PBS_Job ** allJobs = jobCnt?(new PBS_Job *[jobCnt+1]):NULL;
	if(jobCnt)
		{
		memcpy(allJobs, jobs, jobCnt*sizeof(PBS_Job *));
		allJobs[jobCnt] = NULL;
		}
	return allJobs;	
	}

// *****************************************************************************
// * PBS_JobList::getJobsByQueue :
// *	This method returns an array of all jobs in the PBS_Server that are 
// *	currently in a specified queue.  The caller is responsible for 
// *	deleting the array that is returned by this call.
// *****************************************************************************
PBS_Job ** PBS_JobList::getJobsByQueue ( char * queueName )
	{
	PBS_Job ** selectJobs = NULL;
	int        selectCnt  = 0 ;
	size_t     idx;
	
	size_t     maxIndices = 0;
	size_t   * indexList  = NULL;

	for(idx=0; idx<jobCnt; idx++)
		{
		if(jobs[idx] && !strcmp(jobs[idx]->getQueueName(), queueName))
			{
			if(selectCnt>=maxIndices)
				{
				maxIndices=(maxIndices)?(maxIndices*2):256;
				indexList = (size_t *)(indexList?realloc(indexList, maxIndices*sizeof(size_t)):malloc(maxIndices*sizeof(size_t)));
				}
			indexList[selectCnt++] = idx;
			}
		}

	if(selectCnt>0)
		{
		selectJobs = new PBS_Job * [selectCnt+1];

		for(idx=0; idx<selectCnt; idx++)
			{
			selectJobs[idx] = jobs[indexList[idx]];
			}
		
		selectJobs[idx] = NULL;
		free(indexList);
		}

	return selectJobs;
	}

// *****************************************************************************
// * PBS_JobList::getJobsByState :
// *	This method returns an array of all jobs in the PBS_Server that are 
// *	currently in the specified state.  The caller is responsible for 
// *	deleting the array that is returned by this call.
// *****************************************************************************
PBS_Job ** PBS_JobList::getJobsByState ( PBS_Job::JobState state )
	{
	PBS_Job ** selectJobs = NULL;
	int          selectCnt  = 0 ;
	size_t       selectIdx;
	size_t       allIdx;	
	
	for(allIdx=0; allIdx<jobCnt; allIdx++)
		{
		if(jobs[allIdx] && jobs[allIdx]->getState()==state) 
			selectCnt++;
		}

	if(selectCnt>0)
		{
		selectJobs = new PBS_Job * [selectCnt+1];

		for(allIdx=0, selectIdx=0; 
		    allIdx<jobCnt && selectIdx<selectCnt; 
		    allIdx++)
			{
			if(jobs[allIdx] && jobs[allIdx]->getState()==state)
				{
				selectJobs[selectIdx++] = jobs[allIdx];
				}
			}
		selectJobs[selectIdx] = NULL;
		}

	return selectJobs;
	}


// *****************************************************************************
// * PBS_JobList::asciiDump :
// *	Sequentially dumps all of the jobs in the list.
// *****************************************************************************
void PBS_JobList::asciiDump ( FILE * fp )
	{
	for(int i=0; i<jobCnt; i++)
		{
		if(jobs[i]) jobs[i]->asciiDump(fp);
		}
	}
	
