// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	This header defines the PBS_JobRecord class that is used to maintain
//	information within the PBS_ServerDatabase object.
//  Author:  Walt Akers
// -----------------------------------------------------------------------------
#ifndef _JOB_RECORD_H_
#define _JOB_RECORD_H_ 1

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

class PBS_ServerDatabase;
class PBS_JobArray;

// *****************************************************************************
// * class PBS_JobRecord :
// *	This object contains data retrieved from the PBS log files that is 
// *	pertinent to the scheduling process.  The following elements are 
// *	contained by the PBS_JobRecord.
// *
// *	server        : This is the PBS_ServerDatabase object that this PBS_JobRecord 
// *	                is a part of.
// *	queue         : This is a pointer to the queue that this PBS_JobRecord is
// *	                stored in.
// *	owner         : This is a pointer to the PBS_JobArray that contains all
// *	                jobs owned by a specific user.
// *	jobID         : This is the text job identifier for the job. The job is
// *	                stored and retrieved using this value as a key.
// *	jobName       : This is the name that was given to the job by the owner.
// *	timeEnqueued  : This is the time that the job was placed into the queue.
// *	timeStarted   : This is the time that the job began executing.
// *	timeCompleted : This is the time that the job completed executing.
// *	exitStatus    : This is the integer value that the job returned on
// *	                completion.
// *	cpuTimeUsed   : This is the CPU time that the job used - in seconds.
// *	wallTimeUsed  : This is the total runtime for the job - in seconds.
// *	physMemUsed   : This is the amount of physical memory used by the job.
// *	virtMemUsed   : This is the amount of virtual memory used by the job.
// *****************************************************************************
class PBS_JobRecord
{
protected:
	PBS_ServerDatabase & server; 
	PBS_JobArray       * queue;
	PBS_JobArray       * owner;
	char           * jobID;
	char           * jobName;
	time_t           timeEnqueued;
	time_t           timeStarted;
	time_t           timeCompleted;
	int              exitStatus;
	long             cpuTimeUsed;
	long             wallTimeUsed;
	long             physMemUsed;
	long             virtMemUsed;

public:
	PBS_JobRecord      ( PBS_ServerDatabase & Server, const char * JobID );
	~PBS_JobRecord     ( void );

	inline PBS_ServerDatabase & getServer       ( void ) const;
	inline PBS_JobArray       * getQueue        ( void ) const;
	inline PBS_JobArray       * getOwner        ( void ) const;
	inline const char     * getJobID        ( void ) const;
	inline const char     * getJobName      ( void ) const;
	inline time_t           getTimeEnqueued ( void ) const;
	inline time_t           getTimeStarted  ( void ) const;
	inline time_t           getTimeFinished ( void ) const;
	inline int              getExitStatus   ( void ) const;
	inline long             getCPUTimeUsed  ( void ) const;
	inline long             getWallTimeUsed ( void ) const;
	inline long             getPhysMemUsed  ( void ) const;
	inline long             getVirtMemUsed  ( void ) const;

	void                    setQueue        ( const char * Queue );
	void                    setOwner        ( const char * Owner );
	inline void             setJobName      ( const char * JobName );
	inline void             setTimeEnqueued ( time_t TimeEnqueued );
	inline void             setTimeStarted  ( time_t TimeStarted );
	inline void             setTimeFinished ( time_t TimeFinished );
	inline void             setExitStatus   ( int    ExitStatus );
	inline void             setCPUTimeUsed  ( long   CPUTimeUsed );
	inline void             setWallTimeUsed ( long   WallTimeUsed );
	inline void             setPhysMemUsed  ( long   PhysMemUsed );
	inline void             setVirtMemUsed  ( long   VirtMemUsed );      

	void                    asciiDump       ( FILE * fp = stdout, char * prefix = "\0", int prtHeader = 0);
};

// *****************************************************************************
// * PBS_JobRecord::getServer :
// *	Returns a reference to the PBS_ServerDatabase object that owns the job.
// *****************************************************************************
inline PBS_ServerDatabase & PBS_JobRecord::getServer ( void ) const
	{
	return server; 
	}

// *****************************************************************************
// * PBS_JobRecord::getQueue :
// *	Returns a pointer to the PBS_JobArray object that owns the job.
// *****************************************************************************
inline PBS_JobArray * PBS_JobRecord::getQueue ( void ) const
	{
	return queue; 
	}

// *****************************************************************************
// * PBS_JobRecord::getOwner :
// *	Returns a pointer to the PBS_JobArray object that owns the job.
// *****************************************************************************
inline PBS_JobArray * PBS_JobRecord::getOwner ( void ) const
	{
	return owner; 
	}

// *****************************************************************************
// * PBS_JobRecord::getJobID :
// *	Returns the JobID string that uniquely identifies the job.
// *****************************************************************************
inline const char * PBS_JobRecord::getJobID ( void ) const
	{
	return jobID; 
	}

// *****************************************************************************
// * PBS_JobRecord::getJobName :
// *	Returns the user specified name of the job.
// *****************************************************************************
inline const char * PBS_JobRecord::getJobName ( void ) const
	{
	return jobName; 
	}

// *****************************************************************************
// * PBS_JobRecord::getTimeEnqueued :
// *	Returns the time that the job was last enqueued.
// *****************************************************************************
inline time_t PBS_JobRecord::getTimeEnqueued ( void ) const
	{
	return timeEnqueued; 
	}

// *****************************************************************************
// * PBS_JobRecord::getTimeStarted :
// *	Returns the time that the job began execution.
// *****************************************************************************
inline time_t PBS_JobRecord::getTimeStarted ( void ) const
	{
	return timeStarted; 
	}

// *****************************************************************************
// * PBS_JobRecord::getTimeFinished :
// *	Returns the time that the job completed execution.
// *****************************************************************************
inline time_t PBS_JobRecord::getTimeFinished ( void ) const
	{
	return timeCompleted; 
	}
	
// *****************************************************************************
// * PBS_JobRecord::getExitStatus :
// *	Returns the exit status that was generated when the job finished.
// *****************************************************************************
inline int PBS_JobRecord::getExitStatus ( void ) const
	{
	return exitStatus; 
	}

// *****************************************************************************
// * PBS_JobRecord::getCPUTimeUsed :
// *	Returns the total CPU time that the job used during execution.
// *****************************************************************************
inline long PBS_JobRecord::getCPUTimeUsed ( void ) const
	{
	return cpuTimeUsed; 
	}

// *****************************************************************************
// * PBS_JobRecord::getWallTimeUsed :
// *	Returns the total walltime during which the job was executing.
// *****************************************************************************
inline long PBS_JobRecord::getWallTimeUsed ( void ) const
	{
	return wallTimeUsed; 
	}

// *****************************************************************************
// * PBS_JobRecord::getPhysMemUsed :
// *	Returns the total physical memory used by all processes of the job.
// *****************************************************************************
inline long PBS_JobRecord::getPhysMemUsed ( void ) const
	{
	return physMemUsed; 
	}

// *****************************************************************************
// * PBS_JobRecord::getVirtMemUsed :
// *	Returns the total virtual memory used by all processes of the job.
// *****************************************************************************
inline long PBS_JobRecord::getVirtMemUsed ( void ) const
	{
	return virtMemUsed; 
	}

// *****************************************************************************
// * PBS_JobRecord::setJobName :
// *	Allows the caller to set/change the JobName in the PBS_JobRecord object.
// *****************************************************************************
inline void PBS_JobRecord::setJobName ( const char * JobName )
	{
	if(jobName) free (jobName);
	jobName = (JobName && *JobName)?strdup(JobName):NULL;
	}

// *****************************************************************************
// * PBS_JobRecord::setTimeEnqueued :
// *	Allows the caller to set/modify the timeEnqueued for the PBS_JobRecord 
// * 	object.
// *****************************************************************************
inline void PBS_JobRecord::setTimeEnqueued ( time_t TimeEnqueued )
	{
	timeEnqueued = TimeEnqueued;
	}

// *****************************************************************************
// * PBS_JobRecord::setTimeStarted :
// *	Allows the caller to set/change the timeStarted for the PBS_JobRecord.
// *****************************************************************************
inline void PBS_JobRecord::setTimeStarted ( time_t TimeStarted )
	{
	timeStarted = TimeStarted;
	}

// *****************************************************************************
// * PBS_JobRecord::setTimeFinished :
// *	Allows the caller to set/change the timeCompleted for the PBS_JobRecord.
// *****************************************************************************
inline void PBS_JobRecord::setTimeFinished ( time_t TimeFinished )
	{
	timeCompleted=TimeFinished;
	}

// *****************************************************************************
// * PBS_JobRecord::setExitStatus :
// *	Allows the caller to set/change the exitStatus for the PBS_JobRecord.
// *****************************************************************************
inline void PBS_JobRecord::setExitStatus ( int ExitStatus )
	{
	exitStatus=ExitStatus; 
	}

// *****************************************************************************
// * PBS_JobRecord::setCPUTimeUsed :
// *	Allows the caller to set/change the CPUTimeUsed for the PBS_JobRecord.
// *****************************************************************************
inline void PBS_JobRecord::setCPUTimeUsed ( long CPUTimeUsed )
	{
	cpuTimeUsed=CPUTimeUsed;
	}

// *****************************************************************************
// * PBS_JobRecord::setWallTimeUsed :
// *	Allows the caller to set/change the wallTimeUsed for the PBS_JobRecord.
// *****************************************************************************
inline void PBS_JobRecord::setWallTimeUsed ( long WallTimeUsed )
	{
	wallTimeUsed=WallTimeUsed;
	}

// *****************************************************************************
// * PBS_JobRecord::setPhysMemUsed :
// *	Allows the caller to set/change the physMemUsed for the PBS_JobRecord.
// *****************************************************************************
inline void PBS_JobRecord::setPhysMemUsed ( long PhysMemUsed )
	{
	physMemUsed=PhysMemUsed;
	}

// *****************************************************************************
// * PBS_JobRecord::setVirtMemUsed :
// *	Allows the caller to set/change the virtMemUsed for the PBS_JobRecord.
// *****************************************************************************
inline void PBS_JobRecord::setVirtMemUsed ( long VirtMemUsed )
	{
	virtMemUsed=VirtMemUsed;
	}

#endif /* _JOB_RECORD_H_ */
