// *****************************************************************************
// * Copyright (c) 1999      Southeastern Universities Research Association,
// *                         Thomas Jefferson National Accelerator Facility
// * 
// * This software was developed under a United States Government license
// * described in the NOTICE file included as part of this distribution.
// *****************************************************************************
// *
// * File:        PBS_Node.h
// * Author:      Walt Akers
// * Description: This file contains the class definition for the PBS_Node
// *              class.  This class is generally contained by a PBS_Server 
// *              class which provides populates it and provides access to 
// *              its components.
// *****************************************************************************
#ifndef _PBS_NODE_H_
#define _PBS_NODE_H_ 1

#include <PBS_Base.h>

// *****************************************************************************
// * Features requiring communications with the individual MOMs have been
// * removed to improve performance and reliability
// *****************************************************************************
// #include <PBS_MomNode.h>

// *****************************************************************************
// * class PBS_Node :
// *    This class provides the mechanisms required to read all available
// *    status/configuration information from a PBS node.  It is assumed
// *    that this information will be used later for assessing conditions
// *    when scheduling jobs.
// *
// *    The following data items are defined in this class:
// *
// *         name            : The name of the node.
// *         ntype           : The type of node
// *         state           : An enumerated value indicating the state of 
// *                           the node.
// *         keywords        : A list administrator defined strings that 
// *                           describe the node.
// *         jobs            : A list of jobs that are currently executing on 
// *                           the node.
// *         properties      : This is a list of properties and values that 
// *                           are obtained through direct communication with
// *                           the PBS_Node.
// *****************************************************************************
class PBS_Node : public PBS_Base
{
friend class PBS_Server;

public:
	// *********************************************************************
	// * An enumerated list of possible node types.
	// *********************************************************************
	typedef enum
		{
		UNKNOWN_TYPE=0,
		CLUSTER,
		TIMESHARED
		} NodeType;
		
	// *********************************************************************
	// * An enumerated list of possible node states.
	// *********************************************************************
	typedef enum
		{
		UNKNOWN_STATE=0,
		FREE,
		DOWN,
		OFFLINE,
		RESERVED,
		JOB_EXCLUSIVE,
		JOB_SHARING
		} NodeState;
	

protected:

	// *********************************************************************
	// * These are the names associated with the NodeType enumeration
	// *********************************************************************
	static char * NodeTypeList[3];

	// *********************************************************************
	// * These are the names associated with the NodeState enumeration
	// *********************************************************************
	static char * NodeStateList[7];
	
	// *********************************************************************
	// * This is the PBS_Server that created and manages this PBS_Node.
	// *********************************************************************
	class PBS_Server & server;

	// *********************************************************************
	// * Common node attributes.
	// *********************************************************************
	char             * name;
	size_t             np;
	NodeType           ntype;
	NodeState          state;
	char            ** keywords;
	char            ** jobs;
	
	// *********************************************************************
	// * Features requiring communications with the individual MOMs have 
	// * been removed to improve performance and reliability
	// *********************************************************************
	// PBS_MomNode        properties;

	// *********************************************************************
	// * These are loader methods which are called to load data items in
	// * formats that are unique to the PBS_Node object.  Loader methods
	// * for more common formats are defined in the PBS_Base class.
	// *********************************************************************
	void readNodeTypeValue  ( NodeType & val, struct attrl * attrib );
	void readNodeStateValue ( NodeState & val, struct attrl * attrib );

public:
	// *********************************************************************
	// * PBS_Node::PBS_Node :
	// *	This constructor is called with a reference to the PBS_Server
	// *	object and a pointer to a batch_status structure that contains
	// *	configuration information for the node.
	// *	
	// *	This method is protected to prevent it's execution by anyone
	// *	except the PBS_Server object.
	// *********************************************************************	
	PBS_Node ( PBS_Server & Server, struct batch_status * Stat )
		: server             (Server),
		  name               (NULL),
		  np                 (1),
		  ntype              (UNKNOWN_TYPE),
		  state              (UNKNOWN_STATE),
		  keywords           (NULL),
		  jobs               (NULL)
		  // ***********************************************************
		  // * Features requiring communications with the individual
		  // * MOM have been removed to improve reliability and
		  // * performance.
		  // ***********************************************************
		  // properties (Stat->name)
		{
		loadData( Stat );
		}
				

	// *********************************************************************
	// * PBS_Node::~PBS_Node :
	// *	The destructor for the object.  It calls the clear method 
	// *	to release all memory that was allocated during the loadData
	// *    process.
	// *	
	// *	This method is protected to prevent it's execution by anyone
	// *	except the PBS_Server object.
	// *********************************************************************
	~PBS_Node ( void )
		{
		clear();
		}
		
	// *********************************************************************
	// * PBS_Node::loadData :
	// *	Populates the object using the data retrieved from the
	// *	pbs_statnode function. Calls various helper functions to parse 
	// *	and store this information.
	// *	
	// *	This method is protected to prevent it's execution by anyone
	// *	except the PBS_Server object.
	// *********************************************************************
	int  loadData   ( struct batch_status * job );
	
	// *********************************************************************
	// * PBS_Node::clear :
	// *	Frees allocated memory and reinitializes variables to their
	// *	default values.
	// *	
	// *	This method is protected to prevent it's execution by anyone
	// *	except the PBS_Server object.
	// *********************************************************************
	void clear ( void );

public:

	// *********************************************************************
	// * PBS_Node::asciiDump :
	// *	Generates a text dump of the contents of the object.  This 
	// *	output is particularly useful for debugging and reporting the 
	// *	state of the server.
	// *********************************************************************
	void asciiDump  ( FILE * fp = stdout );

	// *********************************************************************
	// * Accessor methods :
	// *	These methods provide access to the data that is stored inside
	// *	the PBS_Node object.
	// *********************************************************************
	inline PBS_Server &  getServer        ( void ) { return server; }
	inline size_t        getNumProcessors ( void ) { return np; }
	inline const char *  getName          ( void ) { return (const char *)name; }
	inline NodeType      getType          ( void ) { return ntype; }
	inline NodeState     getState         ( void ) { return state; }
	inline const char *  getStateName     ( void ) { return NodeStateList[state]; }
	inline const char ** getKeywords      ( void ) { return (const char **)keywords; }
	inline const char ** getJobs          ( void ) { return (const char **)jobs; }

	// *********************************************************************
	// * Features requiring communications with the individual MOMs have 
	// * been removed to improve performance and reliability
	// *********************************************************************
	// inline PBS_MomNode * getProperties    ( void ) { return &properties; }
};

// *****************************************************************************
// * class PBS_NodeList:
// *	This is a managed array of PBS_Node objects.
// *****************************************************************************
class PBS_NodeList 
{
protected:
	PBS_Node ** nodes;
	size_t       nodeCnt;
	
	// *********************************************************************
	// * PBS_NodeList::PBS_NodeList :
	// *	Default constructor for the class.  It initializes all internal
	// *	variables to their default values.
	// *********************************************************************
	PBS_NodeList ( void );

	// *********************************************************************
	// * PBS_NodeList::addNodes :
	// *	This method populates the object using an external array of 
	// *	nodes that is termimated by a NULL pointer.  The array becomes 
	// *	the property of the class and should not be accessed directly 
	// *	by the caller.
	// *********************************************************************
	void addNodes ( PBS_Node ** nodelist );
	
public:
	// *********************************************************************
	// * PBS_NodeList::PBS_NodeList :
	// *	Parameterized constructor for the class.  It initializes the
	// *	internal variables using an array of nodes that is terminated
	// *	by a NULL pointer.
	// *
	// *	The array becomes the property of the object and should not be
	// *	accessed directly by the caller.
	// *********************************************************************
	PBS_NodeList ( PBS_Node ** nodelist );
	
	// *********************************************************************
	// * PBS_NodeList::~PBS_NodeList :
	// *	Destructor for the object.  It calls clear to free the array 
	// *	structure associated with the object. The PBS_Server is 
	// *	responsible for deleteing the individual PBS_Nodes.
	// *********************************************************************
	~PBS_NodeList ( void );

	// *********************************************************************
	// * PBS_NodeList::getNode :
	// *	This method returns a specific node by name.  If the node doesn't
	// *	exist in the list, then a NULL will be returned.
	// *********************************************************************
	PBS_Node * getNode ( char * name );
		
	// *********************************************************************
	// * PBS_NodeList::getNodeCount :
	// *	Returns a count of all nodes that are currently in the list.
	// *********************************************************************
	size_t getNodeCount ( void );

	// *********************************************************************
	// * PBS_NodeList::getNodes :
	// *	This method returns an array of all nodes that are available in 
	// *	the list.  The user is responsible for deleteing the array 
	// *	that is allocated by this call.
	// *********************************************************************
	PBS_Node ** getNodes ( void );

	// *********************************************************************
	// * PBS_NodeList::getNodesByState :
	// *	This method returns an array of all nodes whose state is the 
	// *	same as specified in the NodeState variable.
	// *********************************************************************
	PBS_Node ** getNodesByState ( PBS_Node::NodeState state );
	
	// *********************************************************************
	// * PBS_NodeList::getNodesByType :
	// *	This method returns an array of all nodes whose type is the 
	// *	same as specified in the NodeType variable.
	// *********************************************************************
	PBS_Node ** getNodesByType ( PBS_Node::NodeType type );
	
	// *********************************************************************
	// * PBS_NodeList::asciiDump :
	// *	Sequentially dumps all of the nodes in the list.
	// *********************************************************************
	void asciiDump ( FILE * fp = stdout );
	
	// *********************************************************************
	// * PBS_NodeList::operator [] :
	// *	This method allows the caller to walk through the internal array
	// *	of PBS_Node objects.
	// *********************************************************************
	PBS_Node * operator [] (size_t idx)
		{
		return (idx<nodeCnt)?nodes[idx]:NULL;
		}
};
#endif /* _PBS_NODE_H_ */
