// *****************************************************************************
// * Copyright (c) 1999      Southeastern Universities Research Association,
// *                         Thomas Jefferson National Accelerator Facility
// * 
// * This software was developed under a United States Government license
// * described in the NOTICE file included as part of this distribution.
// *****************************************************************************
// *
// * File:        PBS_Queue.cc
// * Author:      Walt Akers
// * Description: This file contains the source code for the functions defined
// *              in the PBS_Queue class.
// *****************************************************************************
#include <PBS_Queue.h>
#include <PBS_Server.h>
		

// *****************************************************************************
// * PBS_Queue::loadData :
// *	This method can be called by the constructor or independently to cause
// *	the object to be populated with new queue data.
// *
// *	The batch_status object is retrieved by calling the pbs_statque function
// *	using the connection ID of the server.  Each attribute of that object
// *	is then examined and the value is loaded into this class.
// *****************************************************************************
int PBS_Queue::loadData ( struct batch_status * bstat )
	{
	struct attrl        * attrib;

	// *********************************************************************
	// * Call clear to clear the contents of the object.
	// *********************************************************************
	clear();

	// *********************************************************************
	// * Copy the name of the queue into the name variable.
	// *********************************************************************
	name = strdup(bstat->name);

	// *********************************************************************
	// * Load all of the jobs in the server that are in this queue.  Calling
	// * this method here presumes that the PBS_Jobs have already been 
	// * loaded into the server.
	// *********************************************************************
	PBS_JobList::addJobs(server.getJobsByQueue(name));
	
	// *********************************************************************
	// * Process the contents of each of the attributes that was returned
	// * as part of the batch_status object.
	// *********************************************************************
	for(attrib=bstat->attribs; attrib!=NULL; attrib = attrib->next)
		{
		switch(*attrib->name)
			{
			case 'a':
			if     (!strcmp(attrib->name, "acl_host_enable")) readBooleanValue(acl_host_enable, attrib);
			else if(!strcmp(attrib->name, "acl_hosts")) readListValue(acl_hosts, attrib);
			else if(!strcmp(attrib->name, "acl_group_enable")) readBooleanValue(acl_group_enable, attrib);
			else if(!strcmp(attrib->name, "acl_groups")) readListValue(acl_groups, attrib);
			else if(!strcmp(attrib->name, "acl_user_enable")) readBooleanValue(acl_user_enable, attrib);
			else if(!strcmp(attrib->name, "acl_users")) readListValue(acl_users, attrib);
			else if(!strcmp(attrib->name, "alt_router")) readBooleanValue(alt_router, attrib);
			else LogError("PBS_Queue::loadData", "Didn't process attribute %s", attrib->name);
			break;
				
			case 'e':
			if     (!strcmp(attrib->name, "enabled")) readBooleanValue(enabled, attrib);
			else LogError("PBS_Queue::loadData", "Didn't process attribute %s", attrib->name);
			break;
			
			case 'f':
			if     (!strcmp(attrib->name, "from_route_only")) readBooleanValue(from_route_only, attrib);
			else LogError("PBS_Queue::loadData", "Didn't process attribute %s", attrib->name);
			break;
			
			case 'k':
			if     (!strcmp(attrib->name, "kill_delay")) readIntegerValue(kill_delay, attrib);
			else LogError("PBS_Queue::loadData", "Didn't process attribute %s", attrib->name);
			break;
		
			case 'm':
			if     (!strcmp(attrib->name, "max_running")) readIntegerValue(max_running, attrib);
			else if(!strcmp(attrib->name, "max_queuable")) readIntegerValue(max_queuable, attrib);
			else if(!strcmp(attrib->name, "max_user_run")) readIntegerValue(max_user_run, attrib);
			else if(!strcmp(attrib->name, "max_group_run")) readIntegerValue(max_group_run, attrib);
			else LogError("PBS_Queue::loadData", "Didn't process attribute %s", attrib->name);
			break;
		
			case 'P':
			if     (!strcmp(attrib->name, "Priority")) readIntegerValue(priority, attrib);
			else LogError("PBS_Queue::loadData", "Didn't process attribute %s", attrib->name);
			break;
		
			case 'q':
			if     (!strcmp(attrib->name, "queue_type")) readQueueTypeValue(queue_type, attrib);
			else LogError("PBS_Queue::loadData", "Didn't process attribute %s", attrib->name);
			break;
			
			case 'r':
			if     (!strcmp(attrib->name, "resources_default")) readResourceValue(resources_default, attrib);
			else if(!strcmp(attrib->name, "resources_max")) readResourceValue(resources_max, attrib);
			else if(!strcmp(attrib->name, "resources_min")) readResourceValue(resources_min, attrib);
			else if(!strcmp(attrib->name, "resources_assigned")) readResourceValue(resources_assigned, attrib);
			else if(!strcmp(attrib->name, "resources_available")) readResourceValue(resources_available, attrib);
			else if(!strcmp(attrib->name, "route_held_jobs")) readBooleanValue(route_held_jobs, attrib);
			else if(!strcmp(attrib->name, "route_waiting_jobs")) readBooleanValue(route_waiting_jobs, attrib);
			else if(!strcmp(attrib->name, "route_retry_time")) readIntegerValue(route_retry_time, attrib);
			else if(!strcmp(attrib->name, "route_lifetime")) readIntegerValue(route_lifetime, attrib);
			else if(!strcmp(attrib->name, "route_destinations")) readListValue(route_destinations, attrib);
			else LogError("PBS_Queue::loadData", "Didn't process attribute %s", attrib->name);
			break;
		
			case 's':
			if     (!strcmp(attrib->name, "state_count")) readStateCountValue(state_count, attrib);
			else if(!strcmp(attrib->name, "started")) readBooleanValue(started, attrib);
			else LogError("PBS_Queue::loadData", "Didn't process attribute %s", attrib->name);
			break;
		
			case 't':
			if(!strcmp(attrib->name, "total_jobs")) readIntegerValue (total_jobs, attrib);
			else LogError("PBS_Queue::loadData", "Didn't process attribute %s", attrib->name);
			break;
			
			default:
			LogError("PBS_Queue::loadData", "Didn't process attribute %s", attrib->name);
			break;
			}
		}

	return 0;
	}

// *****************************************************************************
// * PBS_Queue::clear :
// *	This method releases all memory that was allocated to the object and 
// *	then restores all internal variables to their default values.  This 
// *	method is generally called as part of object destruction, or to clear
// *	values so they can be repopulated from the server.
// *****************************************************************************
void PBS_Queue::clear ( void )
	{
	nullify(name);
	nullify(acl_hosts);
	nullify(acl_groups);
	nullify(acl_users);
	nullify(route_destinations);
	
	memset(state_count, 0, sizeof(state_count));
	
	queue_type         = EXECUTION;
	enabled            = FALSE;
	started            = FALSE;
	priority           = 0;
	total_jobs         = 0;
	max_running        = 0;
	max_queuable       = 0;
	max_user_run       = 0;
	max_group_run      = 0;
	from_route_only    = FALSE;
	acl_host_enable    = FALSE;
	acl_group_enable   = FALSE;
	acl_user_enable    = FALSE;
	checkpoint_min     = 0;
	kill_delay         = 2;
	alt_router         = FALSE;
	route_held_jobs    = FALSE;
	route_waiting_jobs = FALSE;
	route_retry_time   = 30;
	route_lifetime     = 0;

	resources_default.clear();
	resources_max.clear();
	resources_min.clear();
	resources_assigned.clear();
	resources_available.clear();
	
	PBS_JobList::addJobs(NULL);
	}


// *****************************************************************************
// * PBS_Queue::asciiDump :
// *	This method generates a plain text report of the contents of the
// *	object.  This method is generally used for debugging, however, the
// *	information is also useful for reporting and verifying changes made
// *	from other sources.  
// *****************************************************************************
void PBS_Queue::asciiDump ( FILE * fp )
	{
	int i;

	fprintf(fp, "-----------------------------------------------------\n");
	fprintf(fp, "QUEUE %s\n", name);
	fprintf(fp, "-----------------------------------------------------\n");
	fprintf(fp, "QUEUE CONFIGURATION\n");
	fprintf(fp, "     Server:          %s\n", server.getName());
	fprintf(fp, "     Name:            %s\n", name);
	fprintf(fp, "     Type:            %s\n", queue_type==EXECUTION?"Execution":"Routing");
	fprintf(fp, "     Enabled:         %s\n", enabled?"Yes":"No");
	fprintf(fp, "     Started:         %s\n", started?"Yes":"No");
	fprintf(fp, "     Priority:        %i\n", priority);
	fprintf(fp, "     Kill Delay:      %i seconds\n", kill_delay);
	fprintf(fp, "     From Route Only: %s\n", from_route_only?"Yes":"No");

	fprintf(fp, "\nROUTING QUEUE CONFIGURATION\n");
	fprintf(fp, "     Alt Router:      %s\n", alt_router?"ENABLED":"DISABLED");
	fprintf(fp, "     Route Held:      %s\n", route_held_jobs?"Yes":"No");
	fprintf(fp, "     Route Waiting:   %s\n", route_waiting_jobs?"Yes":"No");
	fprintf(fp, "     Retry Time:      %i\n", route_retry_time);
	fprintf(fp, "     Lifetime:        %i\n", route_lifetime);
	fprintf(fp, "     Destinations:");
	for(i=0; route_destinations && route_destinations[i]; i++) 
		fprintf(fp, "%*c%s\n", (i?22:4), 0x20, route_destinations[i]);
	if(!i) fprintf(fp, "\n");
	
	fprintf(fp, "\nEXECUTION QUEUE CONFIGURATION\n");
	fprintf(fp, "     Max Running:     %i\n", max_running);
	fprintf(fp, "     Max Queuable:    %i\n", max_queuable);
	fprintf(fp, "     Max User Run:    %i\n", max_user_run);
	fprintf(fp, "     Max Group Run:   %i\n", max_group_run);

	fprintf(fp, "\nQUEUE STATISTICS\n");
	fprintf(fp, "     Total Jobs:      %i\n", total_jobs);
	fprintf(fp, "     - In Transit:    %i\n", state_count[TRANSIT]);
	fprintf(fp, "     - Queued:        %i\n", state_count[QUEUED]);
	fprintf(fp, "     - Held:          %i\n", state_count[HELD]);
	fprintf(fp, "     - Waiting:       %i\n", state_count[WAITING]);
	fprintf(fp, "     - Running:       %i\n", state_count[RUNNING]);
	fprintf(fp, "     - Exiting:       %i\n",  state_count[EXITING]);
	
	fprintf(fp, "\nACCESS CONTROL\n");	
	fprintf(fp, "     Host ACL:        %s\n", acl_host_enable?"ENABLED":"DISABLED");
	fprintf(fp, "     Host Names:"); 
	for(i=0; acl_hosts && acl_hosts[i]; i++) 
		fprintf(fp, "%*c%s\n", (i?22:6), 0x20, acl_hosts[i]);
	if(!i) fprintf(fp, "\n");
	
	fprintf(fp, "     Group ACL:       %s\n", acl_group_enable?"ENABLED":"DISABLED");
	fprintf(fp, "     Group Names:");
	for(i=0; acl_groups && acl_groups[i]; i++) 
		fprintf(fp, "%*c%s\n", (i?22:5), 0x20, acl_groups[i]);
	if(!i) fprintf(fp, "\n");
	
	fprintf(fp, "     User ACL:        %s\n", acl_user_enable?"ENABLED":"DISABLED");
	fprintf(fp, "     User Names:");
	for(i=0; acl_users && acl_users[i]; i++) 
		fprintf(fp, "%*c%s\n", (i?22:6), 0x20, acl_users[i]);
	if(!i) fprintf(fp, "\n");

	fprintf(fp, "\nRESOURCES\n");
	fprintf(fp, "     Resource Defaults:\n");
	resources_default.asciiDump(fp, "                      ");
	fprintf(fp, "     Resource Maximums:\n");
	resources_max.asciiDump(fp, "                      ");
	fprintf(fp, "     Resource Minimums:\n");
	resources_min.asciiDump(fp,"                      " );
	fprintf(fp, "     Resources Assigned:\n");
	resources_assigned.asciiDump(fp, "                      ");
	fprintf(fp, "     Resources Available:\n");
	resources_available.asciiDump(fp, "                      ");
	fprintf(fp, "\n");
	fflush(fp);
	}
	
// *****************************************************************************
// * PBS_QueueList::PBS_QueueList :
// *	Default constructor for the class.  It initializes all internal
// *	variables to their default values.
// *****************************************************************************
PBS_QueueList::PBS_QueueList ( void )
	: queues(NULL),
	  queueCnt(0)
	{
	}

// *****************************************************************************
// * PBS_QueueList::PBS_QueueList :
// *	Parameterized constructor for the class.  It initializes the
// *	internal variables using an array of queues that is terminated
// *	by a NULL pointer.
// *
// *	The array becomes the property of the class and should not be
// *	accessed directly by the caller.
// *****************************************************************************
PBS_QueueList::PBS_QueueList ( PBS_Queue ** queuelist )
	: queues  (NULL),
	  queueCnt(0)
	{
	addQueues ( queuelist );
	}

// *****************************************************************************
// * PBS_QueueList::~PBS_QueueList :
// *	Destructor for the object.  It calls clear to free the array structure
// *	associated with the object. The PBS_Server is responsible for 
// *	deleteing the individual PBS_Queues.
// *****************************************************************************
PBS_QueueList::~PBS_QueueList ( void )
	{
	if(queues) free(queues);
	queues   = NULL;
	queueCnt = 0;
	}


// *****************************************************************************
// * PBS_QueueList::addQueues :
// *	This method populates the object using an external array of 
// *	queues that is termimated by a NULL pointer.  The array becomes 
// *	the property of the class and should not be accessed directly 
// *	by the caller.
// *****************************************************************************
void PBS_QueueList::addQueues ( PBS_Queue ** queuelist )
	{
	if(queues) free(queues);
	queues   = queuelist;
	queueCnt = 0;

	if(queues!=NULL) while(queues[queueCnt]!=NULL) queueCnt++;
	}
		
// *****************************************************************************
// * PBS_QueueList::getQueue :
// *	This method returns a specific queue by name.  If the queue doesn't
// *	exist in the list, then a NULL will be returned.
// *****************************************************************************
PBS_Queue * PBS_QueueList::getQueue ( char * name )
	{
	PBS_Queue * result = NULL;
	int idx;
		
	if(name && *name)
		{
		for(idx=0; idx<queueCnt && strcmp(queues[idx]->getName(), name); idx++);
		if(idx<queueCnt) result = queues[idx];
		}
	
	return result;
	}
	
// *****************************************************************************
// * PBS_QueueList::getQueueCount :
// *	Returns a count of all queues that are currently in the list.
// *****************************************************************************
size_t PBS_QueueList::getQueueCount ( void )
	{
	return queueCnt;
	}

// *****************************************************************************
// * PBS_QueueList::getQueues :
// *	This method returns an array of all queues that are available in 
// *	the list.  The user is responsible for deleteing the array 
// *	that is allocated by this call.
// *****************************************************************************
PBS_Queue ** PBS_QueueList::getQueues ( void )
	{
	PBS_Queue ** allQueues = queueCnt?(new PBS_Queue *[queueCnt+1]):NULL;
	if(queueCnt)
		{
		memcpy(allQueues, queues, queueCnt*sizeof(PBS_Queue *));
		allQueues[queueCnt] = NULL;
		}
	return allQueues;	
	}

// *****************************************************************************
// * PBS_QueueList::asciiDump :
// *	Sequentially dumps all of the queues in the list.
// *****************************************************************************
void PBS_QueueList::asciiDump ( FILE * fp )
	{
	for(int i=0; i<queueCnt; i++)
		{
		if(queues[i]) queues[i]->asciiDump(fp);
		}
	}
	


