// *****************************************************************************
// * Copyright (c) 1999      Southeastern Universities Research Association,
// *                         Thomas Jefferson National Accelerator Facility
// * 
// * This software was developed under a United States Government license
// * described in the NOTICE file included as part of this distribution.
// *****************************************************************************
// *
// * File:        PBS_Resource.h
// * Author:      Walt Akers
// * Description: This file contains the class definition for the PBS_Resource
// *              class.  This class is used to collect and store resource
// *              information that is used by numerous PBS objects.
// *****************************************************************************
#ifndef _PBS_RESOURCE_H_
#define _PBS_RESOURCE_H_ 1

#include <stdio.h>
#include <string.h>

// *****************************************************************************
// * class PBS_Resource :
// *	This class provides a simple mechanism for storing and retrieving
// *	resource names and values that are stored on a PBS_Server.
// ***************************************************************************** 
class PBS_Resource
{
protected:
	size_t  rescCnt;	// Number of loaded resources
	size_t  rescMax;	// Current maximum size of resource arrays
	char ** rescName;	// Resource name array
	char ** rescValue;	// Resource value array
	size_t  longestName;	// Longest resource name that has been added
				// (used for report formatting only)
	
public:
	// *********************************************************************
	// * PBS_Resource::PBS_Resource :
	// *	Class constructor - initializes all variables to their default
	// *	values.
	// *********************************************************************
	PBS_Resource ( void )
		: rescCnt(0), 
		  rescMax(0),
		  rescName(NULL),
		  rescValue(NULL),
		  longestName(0)
		{
		}
		
	// *********************************************************************
	// * PBS_Resource::~PBS_Resource :
	// *	Class destructor - releases any memory that was allocated by 
	// *	the class.
	// *********************************************************************
	~PBS_Resource ( void )
		{
		clear();
		if(rescName)  free(rescName);
		if(rescValue) free(rescValue);
		}
		
	// *********************************************************************
	// * PBS_Resource::clear :
	// *	This method deletes all of the name and values that have been
	// *	stored in the object and restores the variables to their 
	// *	initial values.  The arrays (the size of which are specified
	// *	by the rescMax variable) are not deleted because they can be
	// *	reused later.
	// *********************************************************************
	void clear ( void )
		{
		while(rescCnt>0)
			{
			free(rescName[--rescCnt]);
			free(rescValue[rescCnt]);
			}
		}
		
	// *********************************************************************
	// * PBS_Resource::insert :
	// *	This method allows the caller to insert a resource.  The method
	// *	walks through the array until it find a resource with the same
	// *	name -OR- the names are alphabetically higher than the new 
	// *	name.
	// *
	// *	If an entry with the same name is found, then the value will 
	// *	be replaced with the new value.  If no entry is found, then
	// *	all subsequent entries will be moved down and the new entry
	// *	will be inserted into the array.  The size of the array will 
	// *	be increased geometrically each time it is necessary, this 
	// *	should reduce the need for reallocations.
	// *********************************************************************
	void insert ( char * name, char * value )
		{
		int comp;
		int i;
		
		i = (name?strlen(name):0);
		if(i>longestName) longestName = i;
		
		for(i=0; i<rescCnt && (comp=strcmp(name, rescName[i]))>0; i++);
		if(i<rescCnt && comp==0) 
			{
			if(rescValue[i]) free(rescValue[i]);
			rescValue[i] = value?strdup(value):NULL;
			}
		else	{
			if(rescMax < rescCnt+1)
				{
				rescMax = (rescMax<32)?32:rescMax*2;
				rescName = (char **)(rescName?realloc(rescName, rescMax*sizeof(char *)):malloc(rescMax*sizeof(char *)));
				rescValue = (char **)(rescValue?realloc(rescValue, rescMax*sizeof(char *)):malloc(rescMax*sizeof(char *)));
				}
			if((rescCnt-i) > 0)
				{
				memmove(&rescName[i+1],  &rescName[i], (rescCnt-i)*sizeof(char *));
				memmove(&rescValue[i+1], &rescValue[i], (rescCnt-i)*sizeof(char *));
				}
			rescName[i]  = strdup(name);
			rescValue[i] = value?strdup(value):NULL;
			rescCnt++;
			}
		}

	// *********************************************************************
	// * PBS_Resource::remove :
	// *	This method will remove a named resource from the resource
	// *	table.  If the resource specified is not in the class, then 
	// *	no action is taken.  If the resource is found, then it is
	// *	deleted and all subsequent entries are moved up one to ensure
	// *	that  there are no gaps in the array.
	// *********************************************************************
	void remove ( char * name )
		{
		int i;
		int comp;
		
		for(i=0; i<rescCnt && (comp=strcmp(name, rescName[i]))>0; i++);
		if(i<rescCnt && comp==0)
			{
			if(rescName[i]) 
				{
				free(rescName[i]);
				rescName[i] = NULL;
				}
			if(rescValue[i])
				{
				free(rescValue[i]);
				rescValue[i] = NULL;
				}
			if((--rescCnt - i)>0)
				{
				memmove(&rescName[i], &rescName[i+1], (rescCnt-i)*sizeof(char *));
				memmove(&rescValue[i], &rescValue[i+1], (rescCnt-i)*sizeof(char *));
				}
			}
		}

	// *********************************************************************
	// * PBS_Resource::find :
	// *	This method will find a resource with the specified name and 
	// *	return its value.  If the specified name has not been inserted
	// *	in the object, then NULL will be returned.
	// *********************************************************************
	char * find ( char * name )
		{
		int i;
		int comp;
		
		for(i=0; i<rescCnt && (comp=strcmp(name, rescName[i]))>0; i++);
		return (i<rescCnt && comp==0)?rescValue[i]:NULL;
		
		}

	// *********************************************************************
	// * PBS_Resource::getCount :
	// *	This method will return an integer identifying the total number
	// *	of resources name/value pairs stored in the PBS_Resource object.
	// *********************************************************************
	size_t getCount  ( void )
		{
		return rescCnt;
		}
		
	// *********************************************************************
	// * PBS_Resource::getNames :
	// *	This method will return a const pointer to the array of names
	// *	stored in the PBS_Resource object.  Classes that use the
	// *	PBS_Resource object must use caution when using this call 
	// *	and should protect the PBS_Resource object from external access
	// *	to prevent this array from being corrupted. 
	// *********************************************************************
	const char ** getNames ( void ) 
		{
		return (const char **)rescName;
		}
		
	// *********************************************************************
	// * PBS_Resource::getValues :
	// *	This method will return a const pointer to the array of values
	// *	stored in the PBS_Resource object.  Classes that use the
	// *	PBS_Resource object must use caution when using this call 
	// *	and should protect the PBS_Resource object from external access
	// *	to prevent this array from being corrupted. 
	// *********************************************************************
	const char ** getValues ( void )
		{
		return (const char **)rescValue;
		}

	// *********************************************************************
	// * PBS_Resource::asciiDump :
	// *	This method produces an ascii text report of the contents of the
	// *	PBS_Resource object.  The prefix string has been added to allow
	// *	parent classes to install a prefix string to make this objects
	// *	formatting consistent with the owner objects.
	// *********************************************************************
	void asciiDump ( FILE * fp = stdout, char * prefix=NULL)
		{
		for(int i=0; i<rescCnt; i++)
			{
			if(rescName[i] && *rescName[i])
				{
				fprintf(fp, 
					"%s%-*s = %s\n", 
					prefix?prefix:"\0", 
					longestName, rescName[i], 
					rescValue[i]?rescValue[i]:"Undefined");
				}
			}
		fflush(fp);
		}
};

#endif /* _PBS_RESOURCE_H_ */
