// *****************************************************************************
// * Copyright (c) 1999      Southeastern Universities Research Association,
// *                         Thomas Jefferson National Accelerator Facility
// * 
// * This software was developed under a United States Government license
// * described in the NOTICE file included as part of this distribution.
// *****************************************************************************
// *
// * File:        PBS_Server.cc
// * Author:      Walt Akers
// * Description: This file contains the source code for the functions defined
// *              in the PBS_Server class.
// *****************************************************************************
#include <PBS_Server.h>

// *****************************************************************************
// * PBS_Server::ServerStateList :
// * 	An array of strings declared in PBS_Server.h that provide a text
// * 	string name for each of the integer values in the ServerState
// * 	enumeration.
// *****************************************************************************
char * PBS_Server::ServerStateList[5] =
	{
	"Active",
	"Idle",
	"Scheduling",
	"Terminating",
	"Terminating. Delayed"
	};

// *****************************************************************************
// * PBS_Server::loadServerStateValue :
// *	This method is called by the loadData method when the "server_state"
// *	attribute must be processed.  This method parses the value string
// *	and assigns the appropriate integer value to the server_state variable.
// *****************************************************************************
void PBS_Server::readServerStateValue (ServerState & server_state, struct attrl * attrib)
	{
	switch(*attrib->value)
		{
		case 'I':
		server_state = IDLE;
		break;
		
		case 'S':
		server_state = SCHEDULING;
		break;
		
		case 'T':
		if(!strcmp(attrib->value, "Terminating")) server_state=TERMINATING;
		else server_state = DELAYED_TERMINATING;
		break;
		
		default:
		case 'A':
		server_state = ACTIVE;
		break;
		}
	}


// *****************************************************************************
// * PBS_Server::loadData :
// *	This method can be called by the constructor or independently to cause
// *	the object to be populated with new server data.  The method calls the
// *	pbs_statsserver function to obtain configuration data from the PBS
// *	server that is identfied by "connID".  Each of these configuration 
// *	items is then passed to loadValue to be parsed and stored in this
// *    object.
// *****************************************************************************
int PBS_Server::loadData ( int connID )
	{
	int result = 0;
	struct batch_status * bstat;

	// *********************************************************************
	// * Call clear to free any memory that may have been previously
	// * been allocated.
	// *********************************************************************
	clear();

	// *********************************************************************
	// * Call pbs_statserver to retrieve configuration information from the
	// * server.  If the data is read successfully, call loadValue to
	// * sequentially process each string.
	// *********************************************************************
	if(connID>=0 && (bstat = pbs_statserver(connID, NULL, NULL))!=NULL)
		{
		struct attrl *attrib;
		int           i;
		
		// *************************************************************
		// * Record the connection ID and the name of the server.
		// *************************************************************
		connectionID = connID;
		name         = strdup(bstat->name);
		
		// *************************************************************
		// * Process each entry in the batch_status structure and 
		// * stores the value in this object.
		// *************************************************************
		for(attrib=bstat->attribs; attrib!=NULL; attrib = attrib->next)
			{
			switch(*attrib->name)
				{
				case 'a':
				if     (!strcmp(attrib->name, "acl_host_enable")) readBooleanValue(acl_host_enable, attrib);
				else if(!strcmp(attrib->name, "acl_hosts")) readListValue(acl_hosts, attrib);
				else if(!strcmp(attrib->name, "acl_user_enable")) readBooleanValue(acl_user_enable, attrib);
				else if(!strcmp(attrib->name, "acl_users")) readListValue(acl_users, attrib);
				else LogError("PBS_Server::loadData", "Didn't process attribute %s", attrib->name);
				break;
				
				case 'd':
				if     (!strcmp(attrib->name, "default_queue")) readStringValue(default_queue_name, attrib);
				else if(!strcmp(attrib->name, "default_node")) readStringValue(default_node_name, attrib);
				else LogError("PBS_Server::loadData", "Didn't process attribute %s", attrib->name);
				break;
				
				case 'l':
				if     (!strcmp(attrib->name, "log_events")) readIntegerValue(log_events, attrib);
				else LogError("PBS_Server::loadData", "Didn't process attribute %s", attrib->name);
				break;
				
				case 'm':
				if     (!strcmp(attrib->name, "mail_from")) readStringValue(mail_from, attrib);
				else if(!strcmp(attrib->name, "managers")) readListValue(managers, attrib);
				else if(!strcmp(attrib->name, "max_running")) readIntegerValue(max_running, attrib);
				else if(!strcmp(attrib->name, "max_user_run")) readIntegerValue(max_user_run, attrib);
				else if(!strcmp(attrib->name, "max_group_run")) readIntegerValue(max_group_run, attrib);
				else LogError("PBS_Server::loadData", "Didn't process attribute %s", attrib->name);
				break;
				
				case 'n':
				if (!strcmp(attrib->name, "node_pack")) break;
				else LogError("PBS_Server::loadData", "Didn't process attribute %s", attrib->name);

				case 'o':
				if     (!strcmp(attrib->name, "operators")) readListValue(operators, attrib);
				else LogError("PBS_Server::loadData", "Didn't process attribute %s", attrib->name);
				break;
				
				case 'p':
				if     (!strcmp(attrib->name, "pbs_version")) readStringValue(pbs_version, attrib);
				else LogError("PBS_Server::loadData", "Didn't process attribute %s", attrib->name);
				break;
				
				case 'q':
				if (!strcmp(attrib->name, "query_other_jobs")) break;
				else LogError("PBS_Server::loadData", "Didn't process attribute %s", attrib->name);
				break;
				
				case 'r':
				if     (!strcmp(attrib->name, "resources_default")) readResourceValue(resources_default, attrib);
				else if(!strcmp(attrib->name, "resources_max")) readResourceValue(resources_max, attrib);
				else if(!strcmp(attrib->name, "resources_assigned")) readResourceValue(resources_assigned, attrib);
				else if(!strcmp(attrib->name, "resources_available")) readResourceValue(resources_available, attrib);
				else LogError("PBS_Server::loadData", "Didn't process attribute %s", attrib->name);
				break;
				
				case 's':
				if     (!strcmp(attrib->name, "server_state")) readServerStateValue(server_state, attrib);
				else if(!strcmp(attrib->name, "scheduler_iteration")) readIntegerValue(scheduler_iteration, attrib);
				else if(!strcmp(attrib->name, "scheduling")) readBooleanValue (scheduling, attrib);
				else if(!strcmp(attrib->name, "state_count")) readStateCountValue(state_count, attrib);
				else LogError("PBS_Server::loadData", "Didn't process attribute %s", attrib->name);
				break;
				
				case 't':
				if     (!strcmp(attrib->name, "total_jobs")) readIntegerValue (total_jobs, attrib);
				else LogError("PBS_Server::loadData", "Didn't process attribute %s", attrib->name);
				break;
				
				default:
				LogError("PBS_Server::loadData", "Didn't process attribute %s", attrib->name);
				break;
				}
			}
		// *************************************************************
		// * Free any memory allocated to the batch_status structure.
		// *************************************************************
		pbs_statfree(bstat);
		
		// *************************************************************
		// * Call loadNodes to read and process the names and 
		// * configuration of all nodes available to the server.
		// *************************************************************
		loadNodes();
	
		// *************************************************************
		// * Call loadJobs to read and process the names and 
		// * configuration of all jobs on the server.
		// *************************************************************
		loadJobs();
		
		// *************************************************************
		// * Call loadQueues to read and process the names and 
		// * configuration of all queues on the server.
		// *************************************************************
		loadQueues();		
		
		// *************************************************************
		// * Ensure that the loadTime variable is populated here, in the
		// * event that there were no jobs to force the loadTime 
		// * variable to be loaded.
		// *************************************************************
		if(!loadTime) loadTime = time(NULL);		
		}
	else	{
		time_t now = time(NULL);
		if(connID<=0) LogError("PBS_Server::loadData", "Received invalid connection ID");
		else          LogError("PBS_Server::loadData", "pbs_statserver returned NULL");
		result = -1;
		}
	
	return result;
	}


// *****************************************************************************
// * PBS_Server::loadQueues :
// *	This method is called by the loadData method to read the names and 
// *	attributes of all queues that are associated with the server and to
// *	create PBS_Queue objects to store this data.
// *
// *	This method is protected so that it can only be called by the 
// *	PBS_Server::loadData method.
// *****************************************************************************
void PBS_Server::loadQueues ( void )
	{
	struct batch_status * bstat;
	struct attrl        * attrp;
	int                   i;

	// *********************************************************************
	// * Call pbs_statque to retrieve a list of queues that 
	// * are associated with the server.
	// *********************************************************************
	if((bstat = pbs_statque(connectionID, NULL, NULL, NULL))!=NULL)
		{
		struct batch_status * curr = bstat;

		// *************************************************************
		// * Count the number of queues that were retrieved.
		// *************************************************************
		while(curr!=NULL)
			{
			queueCnt++;
			curr = curr->next;
			}

		// *************************************************************
		// * If queues are defined on the server, then allocated
		// * an array to store the associated PBS_Queue objects
		// * and then process the data retrieved from each 
		// * queue.
		// *************************************************************
		if(queueCnt)
			{
			queues = new PBS_Queue *[queueCnt];
			for(i=0, curr=bstat; 
			    i<queueCnt && curr; 
			    i++, curr = curr->next)
				{
				queues[i] = new PBS_Queue(*this, curr);
				}
			}
		
		// *************************************************************
		// * Free the memory allocated to the batch_status structure.
		// *************************************************************
		pbs_statfree(bstat);
		}
	else LogError("PBS_Server::loadQueues", "pbs_statque returned NULL");
	}

// *****************************************************************************
// * PBS_Server::loadJobs :
// *	This method is called by the loadData method to read the names and 
// *	attributes of all jobs that are associated with the server and to
// *	create PBS_Job objects to store this data.
// *
// *	This method is protected so that it can only be called by the 
// *	PBS_Server::loadData method.
// *****************************************************************************
void PBS_Server::loadJobs ( void )
	{
	struct batch_status * bstat;
	struct attrl        * attrp;
	int                   i;

	// *********************************************************************
	// * Call pbs_statque to retrieve a list of jobs that 
	// * are associated with the server.
	// *********************************************************************
	if((bstat = pbs_statjob(connectionID, NULL, NULL, NULL))!=NULL)
		{
		struct batch_status * curr = bstat;

		// *************************************************************
		// * Because the loadTime is critical to determining the start
		// * time of a PBS_Job, populate the loadTime variable here...
		// * It will best reflect the time that the job data was 
		// * extracted from the PBS_Server.
		// *************************************************************
		loadTime = time(NULL);
		
		// *************************************************************
		// * Count the number of jobs that were retrieved.
		// *************************************************************
		while(curr!=NULL)
			{
			jobCnt++;
			curr = curr->next;
			}

		// *************************************************************
		// * If jobs are defined on the server, then allocated
		// * an array to store the associated PBS_Job objects
		// * and then process the data retrieved from each 
		// * job.
		// *************************************************************
		if(jobCnt)
			{
			jobs = new PBS_Job *[jobCnt];
			for(i=0, curr=bstat; 
			    i<jobCnt && curr; 
			    i++, curr = curr->next)
				{
				jobs[i] = new PBS_Job(*this, curr);
				}
			}
		
		// *************************************************************
		// * Free the memory allocated to the batch_status structure.
		// *************************************************************
		pbs_statfree(bstat);
		}
	else LogError("PBS_Server::loadJobs", "pbs_statjob returned NULL");
	}


// *****************************************************************************
// * PBS_Server::loadNodes :
// *	This method is called by the loadData method to read the names and 
// *	attributes of all nodes that are associated with the server and to
// *	create PBS_Node objects to store this data.
// *
// *	This method is protected so that it can only be called by the 
// *	PBS_Server::loadData method.
// *****************************************************************************
void PBS_Server::loadNodes ( void )
	{
	struct batch_status * bstat;
	struct attrl        * attrp;
	int                   i;

	// *********************************************************************
	// * Call pbs_statque to retrieve a list of nodes that 
	// * are associated with the server.
	// *********************************************************************
	if((bstat = pbs_statnode(connectionID, NULL, NULL, NULL))!=NULL)
		{
		struct batch_status * curr = bstat;

		// *************************************************************
		// * Count the number of nodes that were retrieved.
		// *************************************************************
		while(curr!=NULL)
			{
			nodeCnt++;
			curr = curr->next;
			}

		// *************************************************************
		// * If nodes are defined on the server, then allocated
		// * an array to store the associated PBS_Node objects
		// * and then process the data retrieved from each 
		// * node.
		// *************************************************************
		if(nodeCnt)
			{
			nodes = new PBS_Node *[nodeCnt];
			for(i=0, curr=bstat; 
			    i<nodeCnt && curr; 
			    i++, curr = curr->next)
				{
				nodes[i] = new PBS_Node(*this, curr);
				}
			}
		
		// *************************************************************
		// * Free the memory allocated to the batch_status structure.
		// *************************************************************
		pbs_statfree(bstat);
		}
	else LogError("PBS_Server::loadNodes", "pbs_statnode returned NULL");
	}


// *****************************************************************************
// * PBS_Server::clear :
// *	This method releases all memory that was allocated to the object and 
// *	then restores all internal variables to their default values.  This 
// *	method is generally called as part of object destruction, or to clear
// *	values so they can be repopulated from the server.
// *****************************************************************************
void PBS_Server::clear ( void )
	{
	nullify(name);
	nullify(default_queue_name);
	nullify(mail_from);
	nullify(default_node_name);
	nullify(pbs_version);
	nullify(managers);
	nullify(operators);
	nullify(acl_hosts);
	nullify(acl_users);

	memset(state_count, 0, sizeof(state_count));
	
	connectionID    = -1;
	server_state    = ACTIVE;
	scheduling      = FALSE;
	total_jobs      = 0;
	log_events      = 0;
	max_running     = 0;
	max_user_run    = 0;
	max_group_run   = 0;
	acl_host_enable = FALSE;
	acl_user_enable = FALSE;
	loadTime        = 0;
	
	resources_default.clear();
	resources_max.clear();
	resources_assigned.clear();
	resources_available.clear();
	
	if(queues)
		{
		while(queueCnt>0) delete queues[--queueCnt];
		delete[] queues;
		}
	queues   = NULL;
	queueCnt = 0;

	if(nodes)
		{
		while(nodeCnt>0) delete nodes[--nodeCnt];
		delete[] nodes;
		}
	nodes   = NULL;
	nodeCnt = 0;

	if(jobs)
		{
		while(jobCnt>0) delete jobs[--jobCnt];
		delete[] jobs;
		}
	jobs   = NULL;
	jobCnt = 0;
	}

// *****************************************************************************
// * PBS_Server::asciiDump :
// *	This method generates a plain text report of the contents of the
// *	object.  This method is generally used for debugging, however, the
// *	information is also useful for reporting and verifying changes made
// *	from other sources.
// *****************************************************************************
void PBS_Server::asciiDump ( FILE * fp )
	{
	int i;
	fprintf(fp, "-----------------------------------------------------\n");
	fprintf(fp, "SERVER %s\n", name);
	fprintf(fp, "-----------------------------------------------------\n");
	fprintf(fp, "SERVER STATUS\n");
	fprintf(fp, "     Name:          %s\n", name);
	fprintf(fp, "     Connection ID: %i\n", connectionID);
	fprintf(fp, "     State:         %s\n", ServerStateList[server_state]);
	fprintf(fp, "     Scheduling:    %s\n", scheduling?"Yes":"No");
	fprintf(fp, "     Total Jobs:    %i\n", total_jobs);
	fprintf(fp, "     - In Transit:  %i\n", state_count[TRANSIT]);
	fprintf(fp, "     - Queued:      %i\n", state_count[QUEUED]);
	fprintf(fp, "     - Held:        %i\n", state_count[HELD]);
	fprintf(fp, "     - Waiting:     %i\n", state_count[WAITING]);
	fprintf(fp, "     - Running:     %i\n", state_count[RUNNING]);
	fprintf(fp, "     - Exiting:     %i\n", state_count[EXITING]);

	fprintf(fp, "\nSERVER CONFIGURATION\n");
	fprintf(fp, "     PBS Version:   %s\n", pbs_version);
	fprintf(fp, "     Total Queues:  %i\n", queueCnt);
	fprintf(fp, "     Default Node:  %s\n", default_node_name);
	fprintf(fp, "     Default Queue: %s\n", default_queue_name);
	fprintf(fp, "     Mail From:     %s\n", mail_from);
	fprintf(fp, "     Sched Period:  %i\n", scheduler_iteration);
	fprintf(fp, "     Managers:");
	for(i=0; managers && managers[i]; i++) 
		fprintf(fp, "%*c%s\n", (i?20:6), 0x20, managers[i]);
	if(!i) fprintf(fp, "\n");
	fprintf(fp, "     Operators:");
	for(i=0; operators && operators[i]; i++) 
		fprintf(fp, "%*c%s\n", (i?20:5), 0x020, operators[i]);
	if(!i) fprintf(fp, "\n");
	fprintf(fp, "     Log Events:    Code %i%s%s%s%s%s%s%s%s%s\n", log_events,
		(log_events&1)  ?"\n                    - Internal PBS Errors":"\0",
		(log_events&2)  ?"\n                    - Operating System Errors":"\0",
		(log_events&4)  ?"\n                    - Administrative Events":"\0",
		(log_events&8)  ?"\n                    - Job Events":"\0",
		(log_events&16) ?"\n                    - Job Resource Usage":"\0",
		(log_events&32) ?"\n                    - Security Events/Violations":"\0",
		(log_events&64) ?"\n                    - Scheduler Events":"\0",
		(log_events&128)?"\n                    - Common Debugging Output":"\0",
		(log_events&256)?"\n                    - Less Common Debugging Output":"\0");

	fprintf(fp, "\nJOB LIMITS\n");
	fprintf(fp, "     Max Running:   %i\n", max_running);
	fprintf(fp, "     Max User Run:  %i\n", max_user_run);
	fprintf(fp, "     Max Group Run: %i\n\n", max_group_run);

	fprintf(fp, "\nACCESS CONTROL\n");	
	fprintf(fp, "     Host ACL:      %s\n", acl_host_enable?"ENABLED":"DISABLED");
	fprintf(fp, "     Host Names:"); 
	for(i=0; acl_hosts && acl_hosts[i]; i++) 
		fprintf(fp, "%*c%s\n", (i?20:4), 0x20, acl_hosts[i]);
	if(!i) fprintf(fp, "\n");
	fprintf(fp, "     User ACL:      %s\n", acl_user_enable?"ENABLED":"DISABLED");
	fprintf(fp, "     User Names:");
	for(i=0; acl_users && acl_users[i]; i++) 
		fprintf(fp, "%*c%s\n", (i?20:4), 0x20, acl_users[i]);
	if(!i) fprintf(fp, "\n");

	fprintf(fp, "\nRESOURCES\n");
	fprintf(fp, "     Resource Defaults:\n");
	resources_default.asciiDump(fp, "                    ");
	fprintf(fp, "     Resource Maximums:\n");
	resources_max.asciiDump(fp, "                    ");
	fprintf(fp, "     Resources Assigned:\n");
	resources_assigned.asciiDump(fp, "                    ");
	fprintf(fp, "     Resources Available:\n");
	resources_available.asciiDump(fp, "                    ");
	fprintf(fp, "\n");
	fflush(fp);
	
	PBS_QueueList::asciiDump(fp);
	PBS_JobList::asciiDump(fp);
	PBS_NodeList::asciiDump(fp);
	
	fprintf(fp, "-----------------------------------------------------\n");
	fprintf(fp, "END OF REPORT FOR SERVER %s\n", name);
	fprintf(fp, "-----------------------------------------------------\n");
	fflush(fp);
	}
