// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
//
//  Description:
// 	FreeProcMatrix.cc contains the definitions for the methods defined
// 	in the FreeProcMatrix.h file for the classes used to generate and 
// 	sort shared periods of availability on a node.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------
#include <FreeProcMatrix.h>	// Needed for local class definitions
#include <SimpleProcessor.h>	// Needed for class SimpleProcessor
#include <FreeTimeSet.h>	// Needed for class FreeTimeSet
#include <PBS_Log.h>		// Needed for LogError
#include <string.h>		// Needed for memset

// *****************************************************************************
// * FreeProcMatrix::FreeProcMatrix :
// *	This is the constructor for the FreeProcMatrix class. It initializes
// *	the internal variables and allocates a defcault starting number
// *	of FreeProcPeriod objects.
// *****************************************************************************
FreeProcMatrix::FreeProcMatrix ( void )
	: procs      (NULL),
	  head       (NULL),
	  procCnt    (0),
	  procMax    (32),
	  resortFlag (1),
	  relinkFlag (1)
	{
	procs = (FreeProcPeriod *)malloc(procMax*sizeof(FreeProcPeriod));
	clear();
	}
	
// *****************************************************************************
// * FreeProcMatrix::~FreeProcMatrix:
// *	This is the destructor for the FreeProcMatrix object.  It will 
// *	delete all of the FreeProcPeriod objects that were allocated
// *	during its operation.
// *****************************************************************************
FreeProcMatrix::~FreeProcMatrix ( void )
	{
	free(procs);
	}

// *****************************************************************************
// * FreeProcMatrix::clear :
// *	This method sets the procCnt flag to 0 and clears the contents 
// *	of all allocated FreeProcPeriod objects.  This method is 
// *	performed in expectation of repopulating the object with new 
// *	data.
// *****************************************************************************
void FreeProcMatrix::clear ( void )
	{
	procCnt    = 0;
	resortFlag = 1;
	relinkFlag = 1;
	head       = NULL;
	memset(procs, 0, procMax*sizeof(FreeProcPeriod));
	}
	
// *****************************************************************************
// * FreeProcMatrix::detach :
// *	This method will detach a pointer to a FreeProcPeriod from the linked
// *	list that is managed by this object.  The period object will not be 
// *	removed from the ordered array and will be restored to the linked list
// *	the next time the link method is executed.
// *****************************************************************************
void FreeProcMatrix::detach ( FreeProcPeriod * period )
	{
	if(period)
		{
		if(period==head) head = period->next;
		if(period->prev) period->prev->next = period->next;
		if(period->next) period->next->prev = period->prev;
		period->prev = NULL;
		period->next = NULL;

		relinkFlag = 1;
		}
	}
	

// *****************************************************************************
// * FreeProcMatrix::append:
// *	This method allows the caller to dynamically create and insert
// *	a new FreeProcPeriod object into the list.
// *****************************************************************************
void FreeProcMatrix::append ( SimpleProcessor * Proc, FreeTimeSlot * Slot )
	{
	if(Proc && Slot)
		{
		if(procCnt>=procMax)
			{
			while(procCnt>=procMax) procMax*=2;
			procs = (FreeProcPeriod *)realloc(procs, procMax*sizeof(FreeProcPeriod));
			memset (&procs[procCnt], 0, (procMax-procCnt)*sizeof(FreeProcPeriod));
			}

		procs[procCnt].proc   = Proc;
		procs[procCnt++].slot = Slot;

		resortFlag=(relinkFlag = 1);
		}
	}

// *****************************************************************************
// * FreeProcMatrix::append :
// *	This method will append all of the FreeTimeSlot objects that 
// *	are defined within a SimpleProcessor.
// *****************************************************************************
void FreeProcMatrix::append ( SimpleProcessor * Proc )
	{
	if(Proc)
		{
		FreeTimeSlot * slot = Proc->getSlots();
		while(slot!=NULL) 
			{
			append(Proc, slot);
			slot = slot->getNext();
			}
		}
	}
	

// *****************************************************************************
// * FreeProcMatrix::comp :
// * 	This is the comparison function that is used by sort. It compares
// *	two FreeProcPeriod objects.
// *
// *	0  - The two objects are equal.
// *	-1 - a is earlier/longer than b
// *	1  - a is later/shorter than b
// *****************************************************************************
int FreeProcMatrix::comp ( const void * s1, const void * s2 )
	{
	long result = 0;
	
	if(s1 && s2)
		{
		FreeTimeSlot & a = *((FreeProcPeriod *)s1)->slot;
		FreeTimeSlot & b = *((FreeProcPeriod *)s2)->slot;

		if((result = a.getStartTime()-b.getStartTime())==0)
			result = b.getDuration()-a.getDuration();
		}
	else if (s1) result = -1;
	else if (s2) result =  1;
	
	return (result<0)?(-1):((result>0)?1:0);
	}
	
// *****************************************************************************
// * FreeProcMatrix::sort : 
// *	This method calls qsort to resort the array.
// *****************************************************************************
void FreeProcMatrix::sort ( void )
	{
	if(resortFlag && procCnt && procs)
		{
		qsort(procs, procCnt, sizeof(FreeProcPeriod), FreeProcMatrix::comp);
		resortFlag = 0;
		relinkFlag = 1;
		}
	}


// *****************************************************************************
// * FreeProcMatrix::link :
// *	This method is called to relink the sorted array of FreeProcPeriod 
// *	objects.
// *****************************************************************************
void FreeProcMatrix::link ( void )
	{
	if(procCnt && procs)
		{
		if(resortFlag) sort();
		if(relinkFlag)
			{
			head = procs;
			for(int idx=0; idx<procCnt; idx++)
				{
				procs[idx].prev = (idx>0)?&procs[idx-1]:NULL;
				procs[idx].next = (idx<(procCnt-1))?&procs[idx+1]:NULL; 
				}
			relinkFlag = 0;
			}
		}
	else head = NULL;
	}

// *****************************************************************************
// * FreeProcMatrix::merge :
// *	This method will call the sort routine to sort the regions of free time
// *	by startTime and duration.  It will then accumulate these regions into 
// *	periods of free time that span 'size' processors.  A pointer to a list 
// *	of FreeTimeSet objects will be returned.
// *
// * Note that the FreeTimeSet objects generated by this method become
// * the property of the caller and should be deleted when they are no 
// * longer needed.
// *****************************************************************************
FreeTimeSet * FreeProcMatrix::merge ( size_t size )
	{
	// *********************************************************************
	// * If a processor count of 0 is requested or there are no processors,
	// * return NULL.
	// *********************************************************************
	if(size==0 || procCnt==0) return NULL;
	
	// *********************************************************************
	// * If the caller has requested a number of processors that is in 
	// * excess of the allowed processors per node - then report an error
	// * and return NULL.
	// *********************************************************************
	if(size>SchedStatic::MAXPROC)
		{
		LogError("FreeProcMatrix", 
			 "User requests %i procs/node, SchedStatic::MAXPROC is %i",
			 size, SchedStatic::MAXPROC);
		return NULL;
		}
	
	// *********************************************************************
	// * In order to get an accurate representation of the combinations that 
	// * are available, the list must be correctly sorted and linked prior 
	// * to starting...  
	// *********************************************************************
	if(resortFlag) sort();
	if(relinkFlag) link();
	
	// *********************************************************************
	// * These are pointers to two FreeTimeSet objects.  The 'first' object
	// * is at the head of the list and the 'last' object is at the tail.
	// * These objects will be populated with all of the unique multi-
	// * processor combinations that are generated by the merge method.
	// * When the process is finished, the 'first' pointer will be returned
	// * to the caller.  
	// *
	// * Note that the FreeTimeSet objects generated by this method become
	// * the property of the caller and should be deleted when they are no 
	// * longer needed.
	// *********************************************************************
	FreeTimeSet * first = NULL;
	FreeTimeSet * last  = NULL;
	
	// *********************************************************************
	// * Declare and allocate variables that must be consistent across
	// * multiple loop iterations.
	// *********************************************************************
	time_t nextStart = head?head->slot->getStartTime():0;
	SimpleProcessor ** procList = new SimpleProcessor *[size];	

	// *********************************************************************
	// * This method will consume FreeTimeSlots and use them to create
	// * FreeTimeSets until there are no more available.  When the head
	// * variable becomes NULL, then all of the FreeTimeSlots have been
	// * processed.
	// *********************************************************************
	while(head!=NULL)
		{
		FreeProcPeriod * curr      = head;
		FreeProcPeriod * pSave     = NULL;
		time_t           setStart  = curr->slot->getStartTime();
		time_t           setEnd    = curr->slot->getEndTime();
		size_t           procCount = 0;	
		
		// *************************************************************
		// * Copy the address of the SimpleProcessor in the current
		// * FreeProcPeriod to the procList array.
		// *************************************************************
		procList[procCount++] = curr->proc;
		
		// *************************************************************
		// * While the required number of processors have not been 
		// * collected and there continues to be FreeProcPeriods to 
		// * evaluate.
		// *************************************************************
		while(	procCount<size && (curr=curr->next)!=NULL && 
			(nextStart=curr->slot->getStartTime())<setEnd)
			{
			// *****************************************************
			// * Copy the startTime of the next FreeProcPeriod into
			// * the nextStart variable.
			// *****************************************************
			if(curr->next!=NULL) 
				nextStart = curr->next->slot->getStartTime();

			// *****************************************************
			// * Copy the startTime of the current FreeProcPeriod
			// * to the setStart variable.
			// *****************************************************
			setStart = curr->slot->getStartTime();
			
			// *****************************************************
			// * If the endTime of the current FreeProcPeriod is 
			// * less than the recorded setEnd, then update the 
			// * setEnd variable to reflect the new time.
			// *****************************************************
			if(curr->slot->getEndTime()<setEnd)
				setEnd = curr->slot->getEndTime();
				
			// *****************************************************
			// * Increment procCount to indicate that another 
			// * processor has been added to the collection.
			// *****************************************************
			procList[procCount++] = curr->proc;
			}

		// *************************************************************
		// * Once overlapping time from 'size' processors have been 
		// * collected, it is time to create a FreeTimeSet object.
		// *************************************************************
		if(procCount==size)
			{
			// *****************************************************
			// * Compare the time period created with the last time
			// * period on the list.  If the new time period is 
			// * completely encapsulated by an existing one, then 
			// * it will not be necessary to add it to the list ---
			// * It would be redundant.
			// *
			// * Note: we are also testing to ensure that the length
			// * of the FreeTimeSet generated is longer than the 
			// * minimum duration specified in the SchedStatic class
			// *****************************************************
			if((setEnd-setStart)>SchedStatic::getFreeSetMin() &&
			   (!last || last->getEndTime()<setEnd))
				{
				FreeTimeSet * newSet = 
					new FreeTimeSet(procList, size, 
							setStart, setEnd-setStart);

				if(last) last->setNext(newSet);
				else     first = newSet;
				
				last = newSet;
				}
			// *****************************************************
			// * At this point it is possible to elliminate several
			// * objects that have the potential for generating
			// * useless, redundant entries...  Walk forward in the
			// * list until a FreeProcPeriod has a startTime greater
			// * than the setEnd variable.
			// ***************************************************** 
			while(curr->next && curr->slot->getStartTime()<setEnd)
				{
				curr = curr->next;
				}
			
			// *****************************************************
			// * Walk backward through the list and detach any 
			// * FreeProcPeriod object that ends prior to the time
			// * specified in the setEnd value --- we know that
			// * these entries will not be able to generate a unique
			// * FreeTimeSet object.
			// *****************************************************
			while(curr!=NULL)
				{
				pSave = curr->prev;
				if(curr->slot->getEndTime()<=setEnd)
					{
					detach(curr);
					}
				curr = pSave;
				}
			}
		// *************************************************************
		// * If the list of FreeProcPeriod objects ran out before a
		// * sufficient number of processors could be found, then set
		// * the head variable to NULL to prevent further inspection of
		// * the list.
		// *************************************************************
		else if (curr==NULL) head = NULL;
		// *************************************************************
		// * In any other case, we know that the processing loop was
		// * terminated because the start of the current FreeProcPeriod
		// * begins outside of the range that has been garnered in the
		// * setStart and setEnd variables.  Consequently, a quorum of
		// * processors can not be generated from the existing 
		// * FreeProcPeriod objects.  Walk backward through the list
		// * and elliminate any FreeProcPeriod object that ends before
		// * the time specified in nextStart.
		// *************************************************************
		else while(curr!=NULL)
			{
			pSave = curr->prev;
			if(curr->slot->getEndTime()<nextStart) detach(curr);
			curr = pSave;
			}
		}

	// *********************************************************************
	// * Always mark the relinkFlag following the merge method.
	// *********************************************************************
	relinkFlag = 1;
	
	// *********************************************************************
	// * Delete the procList array that was temporarily allocated to store
	// * the address of processors.
	// *********************************************************************
	delete[] procList;
	
	// *********************************************************************
	// * Return the 'first' FreeTimeSet object to the caller.  Note that the
	// * caller is responsible for deleteing each of the FreeTimeSet objects
	// * that have been allocated to this list.
	// *********************************************************************
	return first;
	}

// *****************************************************************************
// * FreeProcMatrix::asciiDump :
// *	This method is used to generate an ascii diganostics report of the 
// *	contents of the object. 
// *****************************************************************************
void FreeProcMatrix::asciiDump ( FILE * fp , char * prefix )
	{
	char c = 0;
	char * prefixString = prefix?prefix:&c;
	
	for(int idx=0; idx<procCnt; idx++)
		{
		fprintf(fp, "%s%s available from %s", 
			prefixString, procs[idx].proc->getName(), 
			SchedStatic::getTimeString(procs[idx].slot->getStartTime()));
		fprintf(fp, " to %s\n", 
			SchedStatic::getTimeString(procs[idx].slot->getEndTime()));
		}	
	}
	
