// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	The FreeTimeSlot.h file contains the header definitions for the 
// 	FreeTimeSlot class.  This class is used to maintain a list of free
// 	time intervals on an individual processor.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------
#ifndef _FREE_TIME_SLOT_H_
#define _FREE_TIME_SLOT_H_ 1

#include <stdio.h>		// Needed for FILE and stdout
#include <time.h>		// Needed for time_t

// *****************************************************************************
// * class FreeTimeSlot :
// *	This class maintains a list of free time intervals that exist on 
// *	an individual processor.  The FreeTimeSlot object is used to ascertain
// *	periods of availability for scheduling jobs.
// *
// *	The data maintained by the FreeTimeSlot object is quite fundamental - 
// *	a starttime for the free time interval and its duration.
// *
// *	This object also maintains a static freeList that is used to keep
// *	a list of preallocated FreeTimeSlot objects.  Whenever a new object 
// *	of this type is requested, the class will attempt to provide one from
// *	the preallocated list in order to reduce memory allocations and improve
// *	overall performance.
// *****************************************************************************
class FreeTimeSlot
{
private:
	// *********************************************************************
	// * The freeList pointer is used to maintain a list of preallocated
	// * FreeTimeSlot objects that may be provided to the requestor.  This
	// * approach reduces the number of dynamic allocations and improves
	// * overall performance.
	// *********************************************************************
	static FreeTimeSlot * freeList;

	FreeTimeSlot * next;
	time_t         startTime;
	time_t         duration;
	
	// *********************************************************************
	// * This is the private constructor that is used by the freeList to 
	// * generate an array of uninitialized FreeTimeSlot objects.
	// *********************************************************************
	FreeTimeSlot ( void );
	
public:
	// *********************************************************************
	// * These are the new and delete operators used to manage the freelist
	// * and provide the caller with a preallocated FreeTimeSlot object
	// * whenever possible.
	// *********************************************************************
	void * operator new ( size_t size );
	void operator delete ( void * ptr );
	
	// *********************************************************************
	// * These are the public constructor and destructor.
	// *********************************************************************
	FreeTimeSlot ( time_t StartTime, time_t Duration );
	~FreeTimeSlot ( void );
	
	// *********************************************************************
	// * These are the inline accessor methods.
	// *********************************************************************
	FreeTimeSlot * getNext      ( void ) { return next; }
	time_t         getStartTime ( void ) { return startTime; }
	time_t         getEndTime   ( void ) { return startTime+duration; }
	time_t         getDuration  ( void ) { return duration; }
	void           setNext      ( FreeTimeSlot * Next ) { next = Next; }
	void           setStartTime ( time_t StartTime )    { startTime = StartTime; }
	void           setDuration  ( time_t Duration )     { duration = Duration; }
  
	// *********************************************************************
	// * This is the ascii reporting method used to generate a text dump of
	// * the contents of the object for diagnostic purposes.
	// *********************************************************************
	void asciiDump ( FILE * fp = stdout, char * prefix = NULL );
};

#endif /* _FREE_TIME_SLOT_H_ */
