// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
//	RangeList.h contains the class definition for the methods and attributes
// 	of the RangeList class.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#ifndef _RANGE_LIST_H_
#define _RANGE_LIST_H_ 1

#include <RangeDefinition.h>		// Needed to inherit from base class

// *****************************************************************************
// * Class declarations for classes that will be used or referenced locally. By
// * minimizing the inclusion of header files, many external dependencies can
// * be elliminated.
// *****************************************************************************
class Range;
class FreeTimeSet;

// *****************************************************************************
// * class RangeList :
// *	The RangeList class is a container for a collection of Range objects
// *	that define periods of availability as compared to the duration of a
// *	specific job.  The Range objects that are contained within the RangeList
// *	contain a rangeStart and rangeDuration. While the rangeStart value
// *	is still a time_t indicating a startTime, the rangeDuration variable
// *	represents a period of time in which a job of the specified type CAN
// *	BE STARTED and still have sufficient available free time to complete.
// *	In other words, a window of opportunity.
// *****************************************************************************
class RangeList : public RangeDefinition 
{
private:
	// *********************************************************************
	// * RangeList * next :
	// *	This next pointer is used to install the RangeList object
	// *	(representing the ranges on a single node) into a ClusterRange
	// *	object that represents the ranges on the entire cluster.
	// *********************************************************************
	RangeList * next;
	
	// *********************************************************************
	// * ClusterNode & node;
	// *	This is a reference to the ClusterNode object where the 
	// *	available processors reside.
	// *********************************************************************
	ClusterNode & node;
	
	// *********************************************************************
	// * Range * range;
	// *	This is an ordered {by startTime}, linked list of time ranges 
	// *	when the ClusterNode will be able to provide the number of 
	// *	processors that the specified job requires.
	// *********************************************************************
	Range * range;

public:
	// *********************************************************************
	// * RangeList::RangeList :
	// *	This is the constructor for the RangeList object.  It is 
	// *	responsible for initializing the internal variables.  The user
	// *	must provide a reference to the owning ClusterNode object, the
	// *	duration, number of nodes requested, number of processors-per-	
	// *	node requested and a PBS_Keywordtable that can be used to 
	// *	initialize the underlying RangeDefinition object.
	// *********************************************************************
	RangeList ( ClusterNode &, time_t, size_t, size_t, PBS_KeywordTable & );

	// *********************************************************************
	// * RangeList::RangeList :
	// *	This is an alternate form of constructor.  It can be called 
	// *	using only a reference to the ClusterNode object and a 
	// *	reference to a RangeDefinition object.
	// *********************************************************************
	RangeList ( ClusterNode &, RangeDefinition & );
	
	// *********************************************************************
	// * RangeList::~RangeList :
	// *	This is the destructor for the RangeList class. It will delete
	// *	any Range objects that have allocated during its lifetime.
	// *********************************************************************
	RangeList::~RangeList ( void );
	
	// *********************************************************************
	// * RangeList::addRanges :
	// *	This method will walk through a linked list of FreeTimeSet 
	// *	objects and will add all FreeTimeSet objects that are 
	// * 	compatible with the jobs requirements.
	// *********************************************************************
	void addRanges ( FreeTimeSet * );
	
	// *********************************************************************
	// * RangeList::addRange :
	// *	This method will create and add a new Range object using the 
	// *	caller provided FreeTimeSet object.
	// *********************************************************************
	void addRange ( FreeTimeSet & );
	
	// *********************************************************************
	// * These are the inline accessor methods that allow the caller to
	// * extract/modify the contents of the object.
	// *********************************************************************
	inline ClusterNode & getNode       ( void )             { return node; }
	inline RangeList   * getNext       ( void )             { return next; }
	inline void          setNext       ( RangeList * Next ) { next = Next; }
	inline Range       * getRange      ( void )             { return range; }
	inline Range       * extractRanges ( void ) 
		{ 
		Range * result=range; 
		range=NULL; 
		return result; 
		}

	// *********************************************************************
	// * RangeList::asciiDump :
	// *	This method generates a diagnostic ascii report of the contents
	// *	of the object.  This data can then be used to evaluate that
	// *	the object is function properly.
	// *********************************************************************
	void asciiDump ( FILE * fp = stdout, char * prefix = NULL );
};

#endif /* _RANGE_LIST_H_ */
