// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	RangeTable.h defines the class attributes and methods for the 
// 	RangeTable class.
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#ifndef _RANGE_TABLE_H_
#define _RANGE_TABLE_H_

#include <stdio.h>		// Needed for FILE structure and stdout
#include <string.h>		// Needed for memset
#include <SchedStatic.h>	// Needed for MAXSUBSET enumeration

// *****************************************************************************
// * Class declarations for classes that will be used or referenced locally. By
// * minimizing the inclusion of header files, many external dependencies can
// * be elliminated.
// *****************************************************************************
class RangeDefinition;
class Range;
class ClusterNode;
class JobCandidate;
class JobDeployment;
class MutualRange;

// *****************************************************************************
// * class RangeTable:
// *	This object is used to collect and manage time ranges that reflect
// *	when a node can execute a job.  These time ranges are combined to 
// *	identify the optimal time for executing a job ( as early as possible).
// *
// *	This class is currently stubbed out until testing is completed on
// *	other objects.
// *****************************************************************************
class RangeTable 
{
protected:

	// *********************************************************************
	// * size_t nTypes :
	// *	This value represents the number of unique node types that are
	// *	requested by a JobCandidate.  Each of the node types is 
	// *	represnted within the JobCandidate by a JobCandidateSubSet
	// *	object.
	// *********************************************************************
	size_t             nTypes;
	
	// *********************************************************************
	// * RangeDefinition ** types :
	// *	The types array contains an array of RangeDefinitions that 
	// *	correspond to the JobCandidateSubSet objects.
	// *********************************************************************
	RangeDefinition * types [ SchedStatic::MAXSUBSETS ];
		
	// *********************************************************************
	// * MutualRange * histogram :
	// *	The histogram is a linked list of MutualRange objects - each of
	// *	which accomodates some or all of the jobs requirements.  The
	// *	histogram will be built until a collection of nodes is located
	// *	that can accomodate all of the jobs requirements... The earliest
	// *	possible, complete MutualRange object will indicate where the
	// *	job will be deployed.
	// *********************************************************************
	MutualRange * histogram;
	
	// *********************************************************************
	// * JobCandidate * job :
	// *	This is the job that is currently being scheduled.
	// *********************************************************************
	JobCandidate * job;
	
public:
	// *********************************************************************
	// * RangeTable::RangeTable :
	// *	This is the constructor for the RangeTable object.  It sets all 
	// *	variables to their default value.
	// *********************************************************************
	RangeTable ( void );
		
	// *********************************************************************
	// * RangeTable::RangeTable :
	// *	This is the constructor for the RangeTable object.  It calls 
	// *	clear to release any information that may have been allocated
	// *	by the object during its use.
	// *********************************************************************
	~RangeTable ( void ) { clear(); }	
	
	// *********************************************************************
	// * These inline accessor methods allow the caller to access the 
	// * contents of the object.
	// *********************************************************************
	inline JobCandidate    * getJob     ( void )     { return job; }
	inline size_t            getTypeCnt ( void )     { return nTypes; }
	inline RangeDefinition * getType    (size_t idx) { return (idx<nTypes?types[idx]:NULL); }

	// *********************************************************************
	// * RangeTable::clear :
	// *	This method is called to clear the RangeTable object in 
	// *	preparation for it to be used to schedule the next job.
	// *********************************************************************
	void clear ( void );

	// *********************************************************************
	// * RangeTable::init :
	// *	This method initializes the RangeTable object with data that is 
	// *	specific to the JobCandidate that is being scheduled.
	// *********************************************************************
	void init ( JobCandidate * );

	// *********************************************************************
	// * RangeTable::compact :
	// *	This method will walk through the MutualRange objects in the
	// *	histogram and will ensure that they are satisfied for all types
	// *	upto and including the type number specified in the typeIdx.
	// *
	// *	If a MutualRange object is not satisfied by the nodes that it 
	// *	has now, then there is no way that it will be satisfied later.
	// *********************************************************************
	void compact ( size_t typeIdx );

	// *********************************************************************
	// * RangeTable::process :
	// *	This method is called with an array of ClusterNode objects that 
	// *	represent the nodes where the job can be scheduled.  This method
	// *	will examine the job requirements and the available nodes and 
	// *	will determine the best time and place for scheduling this job. 
	// *	The method will return a JobDeployment object that reflects the 
	// *	optimal scheduling scenario.
	// *********************************************************************
	JobDeployment * process   ( ClusterNode **, size_t );

	// *********************************************************************
	// * MutualRange * RangeTable::installRange :
	// *	This method is provided to reduce the complexity of the 
	// *	RangeTable::process method.  It is private to the class and is 
	// *	called exclusively by the process method to install an 
	// *	individual range to the correct location in the list of 
	// *	MutualRanges.
	// *
	// *	This method will return the MutualRange object where the 
	// *	allRanges object was installed.
	// *****************************************************************************
	MutualRange * installRange (MutualRange * currPtr, Range & range, size_t type );

	// *********************************************************************
	// * void RangeTable::asciiDump :
	// *	This method generates a diagnostic dump of the RangeTable 
	// *	object.  Since this class is never seen outside of the 
	// *	MatchMatrix object, this method is provided exclusively for 
	// *	internal debugging and testing.
	// *********************************************************************
	void asciiDump ( FILE * fp = stdout, char * prefix = NULL);
};


#endif /* _RANGE_TABLE_H_ */
