// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	File SimpleProcessor.cc realizes the methods that have been declared
// 	for the SimpleProcessor class.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#include <string.h>
#include <SimpleProcessor.h>
#include <JobAssignment.h>

// *****************************************************************************
// * SimpleProcessor::SimpleProcessor :
// *	This is the constructor for the SimpleProcessor object. It will 
// *	initialize all contained objects to their default values and will then
// *	begin processing the user provided parameters.
// *	
// *	The node and index parameters will be combined together to form a
// *	unique processor name of the form "node/index".
// *
// *	A new FreeTimeSlot object will be created that represents the time
// *	from the present (now) until a arbitrary time in the future that is
// *	beyond the life expectancy of any indvidual job.
// *****************************************************************************
SimpleProcessor::SimpleProcessor ( char * node, size_t index )
	: name  (NULL),
	  jobs  (NULL),
	  slots (NULL)
	{
	name = new char [(node?strlen(node):0)+32];
	sprintf(name, "%s/%i", node, index);
	slots = new FreeTimeSlot(SchedStatic::getNow(), 
				 SchedStatic::getTimeMax()-SchedStatic::getNow());
	}
	
// *****************************************************************************
// * SimpleProcessor::~SimpleProcessor :
// *	This is the destructor for the object.  It will delete all job 
// *	assignments and free time slots that were created during its use.
// *****************************************************************************
SimpleProcessor::~SimpleProcessor ( void )
	{
	FreeTimeSlot * slot = slots;
	while(slots!=NULL)
		{
		slot  = slots;
		slots = slot->getNext();
		delete  slot;
		}
	JobAssignment * job = jobs;
	while(jobs!=NULL)
		{
		job  = jobs;
		jobs = job->getNext();
		delete job;
		}
	if(name) delete[] name;
	}
	
// *****************************************************************************
// * SimpleProcessor::getSlot :
// *	This method will locate a free time slot that encompasses the
// *	start time and duration specified as parameters.  If the 
// *	time is not represented, then NULL will be returned.
// *****************************************************************************
FreeTimeSlot  * SimpleProcessor::getSlot  ( time_t start, time_t duration )
	{
	long           diff = 0;
	time_t         end  = start+duration;
	FreeTimeSlot * slot = slots;
	
	while(slot && (diff=(start-slot->getStartTime()))>=0 && slot->getEndTime()<end)
		{
		slot = slot->getNext();
		}

	return (diff>=0)?slot:NULL;
	}
	
	
// *****************************************************************************
// * SimpleProcessor::claim : 
// *	This method is called to install/schedule a job on the SimpleProcessor
// *	object. This method will locate the time interval specified for the
// *	job and will update/split/delete this time interval to reflect the
// *	time utilized by the new job.  The newjob will then be inserted into
// *	the list of JobAssignments. Upon completion, this processors name will
// *	be returned to the calling process to allow it to be registered
// *	elsewhere.
// *****************************************************************************
char * SimpleProcessor::claim ( char * jobName, time_t startTime, time_t duration )
	{
	// *********************************************************************
	// * For the purpose of elliminating obscenely/artificially long jobs
	// * from placing an infinite claim on a processor - we'll institute
	// * the following controls --- no jobs over 1 year.
	// *********************************************************************
	time_t oneYear = 3600*24*365;
	
	if(duration > oneYear) duration=oneYear;
	
	if(startTime<0 || (SchedStatic::getTimeMax()-startTime)<duration)
		{
		startTime = SchedStatic::getTimeMax()-(2*oneYear);
		}
		
	JobAssignment * job = new JobAssignment (jobName, startTime, duration );
	
	// *********************************************************************
	// * The possibility exists that the caller may attempt to claim a node
	// * using a job that has already run beyond its duration before 'now'.
	// * Do not generate a FreeTimeSlot that is earlier than 'now' to
	// * accomodate this attrocity.
	// *********************************************************************
	if(startTime+duration >= SchedStatic::getNow())
		{
		FreeTimeSlot * prevSlot  = NULL;
		FreeTimeSlot * currSlot  = slots;

		// *************************************************************
		// * Walk through the free time slots until you find one that 
		// * ends AFTER the new job starts.
		// *************************************************************
		while(currSlot && currSlot->getEndTime()<startTime)
			{
			prevSlot = currSlot;
			currSlot = currSlot->getNext();
			}	

		// *************************************************************
		// * If an overlapping slot exists - AND IT SHOULD - and it's 
		// * startTime is not an exact match with the startTime of the 
		// * job, split the current time slot in half by performing the 
		// * following steps.
		// *************************************************************
		if (currSlot && currSlot->getStartTime()<startTime)
			{
			// *****************************************************
			// * If the period of time that would be generated on
			// * either side of the job is less than 
			// * SchedStatic::freeSlotMin then it is not necessary
			// * to create a new slot.
			// *****************************************************
			if( startTime-currSlot->getStartTime()<
			    SchedStatic::getFreeSlotMin()) 
			    	{
				currSlot->setDuration(currSlot->getEndTime()-startTime);
				currSlot->setStartTime(startTime);
				}
			 else	{
			 	// *********************************************
				// * Create a second time slot that has the exact 
				// * same start time as the job and the same end 
				// * time as the  current free time slot.
				// *********************************************
				FreeTimeSlot * ns = new FreeTimeSlot
					(startTime, currSlot->getEndTime()-startTime);

				// *********************************************
				// * Attach the FreeTimeSlot that follows the 
				// * current time slot to the new time slot.
				// *********************************************
				ns->setNext(currSlot->getNext());
			
				// *********************************************
				// * Shorten the duration of the current time 
				// * slot so that it ends immediately before the 
				// * new time slot (AND THE JOB) begins.
				// *********************************************
				currSlot->setDuration((ns->getStartTime()-currSlot->getStartTime())-1);
			
				// *********************************************
				// * Insert the new time slot immediately after 
				// * the current time slot.  We have now 
				// * effectively split the current time slot 
				// * into two distinct time slots.
				// *********************************************
				currSlot->setNext(ns);
				prevSlot = currSlot;
				currSlot = ns;
				}
			}
			
		// *************************************************************
		// * Walk forward through the list of FreeTimeSlots and 
		// * elliminate any time slot that is completely encapsulated by 
		// * the job (or is shorter than SchedStatic::freeSlotMin)...
		// *************************************************************
		while(currSlot && currSlot->getEndTime() <= 
		      (job->getEndTime()+SchedStatic::getFreeSlotMin()))
			{
			FreeTimeSlot * nextSlot = currSlot->getNext();
			delete currSlot;
			if(prevSlot) prevSlot->setNext(nextSlot);
			else slots = nextSlot;
			currSlot = nextSlot;
			}
	
		// *************************************************************
		// * At this time currSlot should point to a time slot that 
		// * extends beyond the end of the new job.  If the job and the 
		// * freeTimeSlot overlap, then modify its startTime and 
		// * duration to reflect the presence of the new job.
		// *************************************************************
		if(currSlot && currSlot->getStartTime()<=job->getEndTime())
			{
			currSlot->setDuration ((currSlot->getEndTime()-job->getEndTime())-1);
			currSlot->setStartTime(job->getEndTime()+1);
			}
		}

	JobAssignment * prevJob = NULL;
	JobAssignment * currJob = jobs;
			
	// *********************************************************************
	// * Begin walking through the list of job assignments until you locate
	// * a job that starts after the new job.
	// *********************************************************************
	while(currJob!=NULL && currJob->getStartTime()<=job->getStartTime())
		{
		prevJob = currJob;
		currJob = currJob->getNext();
		}
	
	// *********************************************************************
	// * Insert the new job into the list of job assignments at the 
	// * appropriate location.
	// *********************************************************************
	if(prevJob) prevJob->setNext(job);
	else jobs = job;
	job->setNext(currJob);	
	
	// *********************************************************************
	// * Return the name of this processor so that the calling facility 
	// * can register it elsewhere.
	// *********************************************************************
	return name;
	}
	
// *****************************************************************************
// * SimpleProcessor::scheduled :
// *	This method returns a boolean flag indicating whether the processor has
// *	been scheduled at a specific time.
// *****************************************************************************
int SimpleProcessor::scheduled ( time_t startTime )
	{
	int             result   = 0;
	JobAssignment * job      = jobs;

	while(!result && job!=NULL && job->getStartTime()<=startTime)
		{
		result = (job->getEndTime()>=startTime);
		job = job->getNext();
		}

	return result;
	}

// *****************************************************************************
// * SimpleProcessor::getUtilization :
// *	This method will compute the percentage of time during the specified
// *	time interval that this processor is scheduled for use. A processor that
// *	has a job scheduled for the entire period will have a utilization of 
// *	1.0, while a processor that has no jobs scheduled for the interval will
// *	have a utilization of 0.0.
// *****************************************************************************
double SimpleProcessor::getUtilization ( time_t startTime, time_t duration )
	{
	double result = 0.0;
	
	if(duration>0)
		{
		time_t endTime = startTime+duration;
		JobAssignment * jPtr = jobs;
		
		while(jPtr && jPtr->getEndTime()<startTime) jPtr = jPtr->getNext();
		while(jPtr && jPtr->getStartTime()<endTime)
			{
			double blockStart = 
				(jPtr->getStartTime()>startTime)?
				 jPtr->getStartTime():
				 startTime;
			double blockEnd   =
				(jPtr->getEndTime()<endTime)?
				jPtr->getEndTime():
				endTime;
					
			result += (blockEnd-blockStart)/(double)duration;
			jPtr = jPtr->getNext();
			}
		}
	return result;
	}

	
// *****************************************************************************
// * SimpleProcessor::reserveTime :
// *	This method is called when the system needs to reserve a block
// *	of time to prevent it from being utilized by a job.  This method
// *	will walk through the list of FreeTimeSlot objects and will 
// *	modify any FreeTimeSlot that intersects with the specified
// *	period.
// *****************************************************************************
void SimpleProcessor::reserveTime ( time_t start, time_t duration )
	{
	// *********************************************************************
	// * Make sure the period of reservation does not exceed the maximum
	// * period.
	// *********************************************************************
	if(duration>SchedStatic::getReservationMax()) 
		{
		duration=SchedStatic::getReservationMax();
		}

	time_t         end      = start+duration;
	FreeTimeSlot * prevSlot = NULL;
	FreeTimeSlot * currSlot = slots;	
	
	// *********************************************************************
	// * Walk forward through the list until one of the slots intersects
	// * with the reserved period.
	// *********************************************************************
	while(currSlot && currSlot->getEndTime()<start)
		{
		prevSlot = currSlot;
		currSlot = prevSlot->getNext();
		}
	
	// *********************************************************************
	// * Walk forward through the list until the current slot NO LONGER
	// * intersects with the reserved period.
	// *********************************************************************
	while(currSlot && currSlot->getStartTime()<end)
		{
		FreeTimeSlot * deadSlot = NULL;
		
		if(currSlot->getStartTime()>=start)
			{
			// *****************************************************
			// * Situation A : The FreeTimeSlot is completely 
			// * encapsulated by the reservation period --- Mark
			// * it for deletion.
			// *****************************************************
			if(currSlot->getEndTime()<=end) deadSlot = currSlot;
			// *****************************************************
			// * Situation B : The FreeTimeSlot starts during the 
			// * reserved period and then terminates afterwards --- 
			// * truncate it.
			// *
			// * If the truncated region is LESS THAN the minimum
			// * FreeTimeSlot - then mark it for deletion.
			// *****************************************************
			else	{
				if((currSlot->getEndTime()-end-1)>=SchedStatic::getFreeSlotMin())
					{
					currSlot->setDuration((currSlot->getEndTime()-end)-1);
					currSlot->setStartTime(end+1);
					}
				else deadSlot = currSlot;
				}
			}
		else	{
			// *****************************************************
			// * Situation C : The FreeTimeSlot begins before the 
			// * reserved period and then terminates within it --- 
			// * truncate it.
			// *
			// * If the truncated region is LESS THAN the minimum
			// * FreeTimeSlot - then mark it for deletion.
			// *****************************************************
			if(currSlot->getEndTime()<=end)
				{
				if((start-currSlot->getStartTime()-1)>=SchedStatic::getFreeSlotMin())
					{
					currSlot->setDuration((start-currSlot->getStartTime())-1);
					}
				else deadSlot = currSlot;
				}
			// *****************************************************
			// * Situation D: The FreeTimeSlot starts before the 
			// * reserved period and then ends afterwards --- split 
			// * it.
			// *
			// * Delete either half that has a duration LESS THAN
			// * the minimum FreeTimeSlot. 
			// ******************************************************
			else 	{
				if((currSlot->getEndTime()-end-1)>=SchedStatic::getFreeSlotMin())
					{
					FreeTimeSlot * nextSlot = new FreeTimeSlot(end+1, (currSlot->getEndTime()-end)-1);
					nextSlot->setNext(currSlot->getNext());
					currSlot->setNext(nextSlot);
					}
			
				if((start-currSlot->getStartTime()-1)>=SchedStatic::getFreeSlotMin())
					{
					currSlot->setDuration(start-currSlot->getStartTime()-1);
					}
				else deadSlot = currSlot;
				}
			}
		// *************************************************************
		// * In the end, if the currSlot was either elliminated or 
		// * minimized away, the deadSlot will point to it... 
		// * If deadSlot is non-NULL delete it, otherwise, advance to 
		// * the next object in the list.
		// *************************************************************
		if(deadSlot)
			{
			if(prevSlot) prevSlot->setNext(currSlot->getNext());
			else slots = currSlot->getNext();
			currSlot = currSlot->getNext();
			delete deadSlot;
			}
		else	{
			prevSlot = currSlot;
			currSlot = currSlot->getNext();
			}
		}	
	}
	
// *****************************************************************************
// * SimpleProcessor::asciiDump :
// *	This method is used to generate an ascii report of the contents of the
// * 	processor. This metod may be used for simple reporting or to generate
// *	a diagnostic dump of the object.
// *****************************************************************************
void SimpleProcessor::asciiDump ( FILE * fp, char * prefix )
	{
	char c = 0;
	char * prefixString = prefix?prefix:&c;

	fprintf(fp, "%sProcessor %s\n", prefixString, name );
	fprintf(fp, "%s   Job              Start                End                  Duration \n", prefixString);
	fprintf(fp, "%s  -------------------------------------------------------------------------\n", prefixString);
	JobAssignment * currJob  = jobs;
	FreeTimeSlot  * currSlot = slots;
	
	while(currJob || currSlot)
		{
		if(currSlot && (!currJob || currSlot->getStartTime()<currJob->getStartTime()))
			{
			fprintf(fp, "%s   -- Available --  %-21.21s",
				prefixString,
				SchedStatic::getTimeString(currSlot->getStartTime()));
			fprintf(fp, "%-21.21s",								
				SchedStatic::getTimeString(currSlot->getEndTime()));
			if(currSlot->getEndTime()==SchedStatic::getTimeMax())
				fprintf(fp, "Infinite\n");
			else if(currSlot->getDuration()>=31536000)
				fprintf(fp, "%g years\n", ((double)currSlot->getDuration())/31536000.0);
			else	fprintf(fp, "%04i:%02i:%02i\n", 
					currSlot->getDuration()/3600,
					(currSlot->getDuration()%3600)/60,
					currSlot->getDuration()%60);
			currSlot = currSlot->getNext();
			}
		else if(currJob)
			{
			fprintf(fp, "%s   %-16.16s %-21.21s",
				prefixString, 
				currJob->getName()?currJob->getName():"Unknown",
				SchedStatic::getTimeString(currJob->getStartTime()));
			fprintf(fp, "%-21.21s%04i:%02i:%02i\n",								
				SchedStatic::getTimeString(currJob->getEndTime()),
				currJob->getDuration()/3600,
				(currJob->getDuration()%3600)/60,
				currJob->getDuration()%60);
			currJob = currJob->getNext();
			}
		}
	}
