// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	This file realizes the clas methods that are defined in the 
//	UserPriorityStage.h header file.
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#include <UserPriorityStage.h>

// *****************************************************************************
// * compareFunc :
// *	This is a static comparison method that is used to reorder the jobs
// *	in ascending order by the ownerID.  This will reduce the effort expended
// *	in descending through the list multiple times. This function is static
// *	to prevent it from appearing outside of this object module and to keep
// *	the global name space clean.
// *****************************************************************************
static int compareFunc (const void * a, const void * b)
	{
	PBS_Job * jobA = *(PBS_Job **)a;
	PBS_Job * jobB = *(PBS_Job **)b;
	return jobA->getOwnerID() - jobB->getOwnerID();
	}


// *****************************************************************************
// * UserPriorityStage::compute :
// *	This method is called to compute the weight associated with a job based 
// *	on the user specified priority.  The method will add up all of the 
// *	priorities provided by the jobs owned by a user and then will apply
// *	a weight to each job that represents the priorities percentage of the
// *	sum.
// *****************************************************************************
void UserPriorityStage::compute ( PBS_Job ** jobs, size_t jobCnt, double weightSum )
	{
	size_t idx, uIdx, maxUsers=1;
	double weight = getWeight();

	// *********************************************************************
	// * This structure will be used to accumulate all of the specified 
	// * job priorities and associate them with a user using the jobs'
	// * ownerID. We can preallocate this buffer because their will be no
	// * more than jobCnt users represented.
	// *********************************************************************
	typedef struct
		{
		int id;		// Owner ID
		int least;	// Lowest priority value
		int greatest;	// Greatest priority value
		int adjustment;	// Amounted to make all priorities positive
		} UserPriorityStat;

	// *********************************************************************
	// * Start by getting the weight from the scheduler profile. If the 
	// * weight is 0 or the weightSum is less than or equal to 0, then this
	// * scheduling stage will be skipped. The stage will also be skipped
	// * if the jobCnt is 0 or the jobs arrat is NULL.
	// *********************************************************************
	if(weight<=0.0 || weightSum<=0 || jobCnt==0 || jobs==NULL) return;
		
	// *************************************************************
	// * If a stage debugging recorder has been provided to the
	// * class then use it to record the events associated with this
	// * stage.
	// *************************************************************
	if(recorder)
		{
		recorder->stagePrintf("[User Priority Stage]", "Stage Weight                = %g", weight);
		recorder->stagePrintf("[User Priority Stage]", "Sum of All Stage Weights    = %g", weightSum);
		}
		
	// *********************************************************************
	// * Compute the lower weight boundary.
	// *********************************************************************	
	double LWB = 1.0-(weight/weightSum);

	// *********************************************************************
	// * First - sort the jobs using the ownerID.  This will reduce the
	// * number of times that it is necessary to process the list to 
	// * count, calculate and assign job weights.
	// *********************************************************************
	qsort (jobs, jobCnt, sizeof(PBS_Job *), ::compareFunc);
	
	// *********************************************************************
	// * Allocate an array of UserPriorityStat objects to track information
	// * about user specified priorities.  For uninitialized structures, 
	// * always use memset to clear the contents before proceeding.
	// *********************************************************************
	UserPriorityStat * jStat = new UserPriorityStat[jobCnt];	
	memset(jStat, 0, jobCnt*sizeof(UserPriorityStat));
	
	// *********************************************************************
	// * Walk through the list of jobs and accumulate the priorities 
	// * associated with each ownerID.
	// *********************************************************************
	jStat[0].id = jobs[0]->getOwnerID();
	
	for(idx=0, uIdx=0; idx<jobCnt; idx++)
		{
		if(jStat[uIdx].id != jobs[idx]->getOwnerID())
			{
			jStat[++uIdx].id     = jobs[idx]->getOwnerID();
			jStat[uIdx].least    = jobs[idx]->getPriority();
			jStat[uIdx].greatest = jobs[idx]->getPriority();
			maxUsers++;
			}
		else	{
			if(jStat[uIdx].least > jobs[idx]->getPriority()) 
				jStat[uIdx].least =  jobs[idx]->getPriority();

			if(jStat[uIdx].greatest < jobs[idx]->getPriority())
				jStat[uIdx].greatest = jobs[idx]->getPriority();
			}	
		}

	// *********************************************************************
	// * Walk through the array of registered owner ID's and make 
	// * adjustments if one or more of the user's priorities is less than
	// * 0.
	// *********************************************************************
	for(uIdx=0; uIdx<maxUsers; uIdx++)
		{
		if(jStat[uIdx].least<1)
			{
			jStat[uIdx].adjustment = (1-jStat[uIdx].least);
			jStat[uIdx].greatest  += jStat[uIdx].adjustment;
			}
		else	{
			jStat[uIdx].adjustment = 0;
			}
		}
		
	// *********************************************************************
	// * Walk through the list again, this time applying a weight to each
	// * user's jobs that represents it's priorities perentage of the
	// * summed priorities for that user --- of course, respecting the
	// * weight assigned to the User Priority Stage as a whole.
	// *********************************************************************
	for(idx=0, uIdx=0; idx<jobCnt; idx++)
		{
		if(jStat[uIdx].id != jobs[idx]->getOwnerID()) uIdx++;
		
		double jobWeight =  (double)(jobs[idx]->getPriority()+
				    jStat[uIdx].adjustment)/
				    (double)jStat[uIdx].greatest;
		
		// *************************************************************
		// * Fit the processed weight to the region of the curve that is 
		// * represented by the overall weight of the User Share Stage.
		// *************************************************************
		jobWeight = ((jobWeight+1.0)/2.0)*(1-LWB)+LWB;
		
		// *************************************************************
		// * Obtain the old weight from the job and then apply the 
		// * product of the old weight and this stage's weight as the 
		// * new weight for the job.
		// *************************************************************
		double oldWeight = jobs[idx]->getWeight();
		if(oldWeight<0.0 || oldWeight>1.0) oldWeight = 1.0;
		jobs[idx]->setWeight(oldWeight*jobWeight);

		// *************************************************************
		// * If a stage debugging recorder has been provided to the
		// * class then use it to record the events associated with this
		// * stage.
		// *************************************************************
		if(recorder)
			{
			recorder->jobPrintf(jobs[idx]->getName(), 
				"[User Priority Stage] %s priority %i yields weight %g",
				jobs[idx]->getOwner(),
				jobs[idx]->getPriority(),
				jobWeight);
			}
		}	

	// *********************************************************************
	// * Delete the memory allocated to compute this stage.
	// *********************************************************************
	delete[] jStat;
	}
